"""
Tester.py

GangaRobot IACtion derived class for the TestRobot option:
reads the jobs from the configuration file, and starts a
new instance of the new ganga to test them

"""
from GangaRobot.Framework.Action import IAction
from GangaRobot.Framework import Utility
from Ganga.Utility.logging import getLogger
from Ganga.GPI import *
import Ganga.Utility.Config
import os, datetime, time, shutil
from os.path import join
from GangaRobot.Framework.exceptions import *

logger = getLogger()

class Tester(IAction):

    """ Tester implementation

    Starts up an new pre-release of ganga and tests it

    """

    #class variables

    def execute(self, runid):

        jobsList = []

        config = Ganga.Utility.Config.getConfig('TestRobot')
        OptionList = config['TestPairs']
        InstallDir = config['InstallPath']
        JobDir = config['JobDir']
        ReleaseNo = config['VersionNumber']
        Delta = config['JobTimeOut']
        SleepSecs = 30
      
        i = 0
        for i in range(len(OptionList)):
            TestOption = OptionList[i][0]
            TestBackend = OptionList[i][1]
            TestConfig = OptionList[i][2]
            #make job directory stuff
            Jobfolder = os.path.join(JobDir,ReleaseNo+"_Job_"+str(i))
            try:
                os.mkdir(Jobfolder)
            except OSError as e:
                logger.info(e)
                shutil.rmtree(Jobfolder)
                os.mkdir(Jobfolder)
            #create inifile
            JobIniFile = self._writeinifile(Jobfolder)
            j = Job()
            BackendList = plugins("backends")
            BackendFound = False
            for i in range(len(BackendList)):
                if TestBackend == BackendList[i]:
                    j.backend = BackendList[i]
                    BackendFound = True
                else:
                    pass
            if not BackendFound:
                break # DO SOMETHING ELSE HERE
            from os.path import join
            j.application.exe = join(InstallDir,"install",ReleaseNo,"bin","ganga")
            #j.application.exe = "ganga"            - DEBUGGING LINE
            j.application.args = ["--test","-o[TestingFramework]Config="+TestConfig+".ini",TestOption]
            j.application.env = { 'GANGA_CONFIG_PATH':JobIniFile }
            logger.debug(j.application)
            jobsList += [j]


        i = 0
        self.TestsAllGood = True
        for i in range(len(jobsList)):
            try:
                jobsList[i].submit()
                StartTime = datetime.datetime.now()
                logger.debug("Submitted job "+OptionList[i][0]+" to "+OptionList[i][1]+" backend")
                #Now to wait for time-out
                JobFinished = False
                while not JobFinished:

                    logger.debug("sleeping for %s secs" % str(SleepSecs))
                    time.sleep(SleepSecs)
                    
                    if jobsList[i].status == 'completed':
                        JobFinished = True
                    elif jobsList[i].status == 'failed':
                        JobFinished = True
                    elif (jobsList[i].status == 'running') or (jobsList[i].status == 'submitted'):
                        EndTime = StartTime + datetime.timedelta(seconds=Delta)
                        NowTime = datetime.datetime.now()
                        if (NowTime > EndTime):
                            NowTimeAsString = datetime.datetime.now().strftime("%d %b %Y %H:%M:%S")
                            logger.warning("Job "+str(i)+" has timed out at %s" % NowTimeAsString)
                            jobsList[i].kill()
                            JobFinished = True
                    else:
                        pass
                    # PUT HEARTBEAT HERE
                    try:
                        self._HeartBeat()
                    except:
                        raise GangaRobotFatalError
   
            except:
                logger.error("Failed to submit job "+str(i)+" TestOption %s, TestConfig %s",OptionList[i][0], OptionList[i][2])
                self.TestsAllGood = False

        logger.info("All jobs finished")
        logger.debug(jobs)
        #for j in jobs:
            #j.peek('stdout')
        jobs.remove()
        
        if self.TestsAllGood:
            self._WriteFinalState()
            
        self._cleanup
        
        
    def _writeinifile(self, Jobfolder):
        JobConfigFile = os.path.join(Jobfolder, "TestConfig.ini")
        f = open(JobConfigFile, 'w')
        f.write("[TestingFramework]\n")
        f.write("EnableHTMLReporter = True\n")
        f.write("ReleaseTesting = True\n")
        JobLogOutputDir = os.path.join(Jobfolder,"output")
        f.write("LogOutputDir = "+JobLogOutputDir+"\n")
        f.write("OutputDir = "+Jobfolder+"\n")
        JobReportsOutputDir = os.path.join(Jobfolder,"reports")
        f.write("ReportsOutputDir = "+JobReportsOutputDir+"\n")
        f.write("Config=Imperial.ini")
        f.write("\n")
        f.write("[Configuration]\n")
        f.write("gangadir = "+Jobfolder+"\n")
        f.write("\n")
        f.flush()
        f.close()
        return JobConfigFile
        
    def _cleanup(self):
        # remove installed files.
        pass
        
        
    def _WriteFinalState(self):
        import tempfile, fnmatch
        tmpdir = tempfile.gettempdir()
        config = Ganga.Utility.Config.getConfig('TestRobot')
        CurrentVersionData = config['VersionNumber']+" - "+config['VersionTime']
        gangaconfig = Ganga.Utility.Config.getConfig('Configuration')
        testrobotdir = gangaconfig['gangadir']
        VersionFileFound = False
        from os.path import join,isfile
        lastversionfile = join(str(testrobotdir),"LastVersionFile.txt")
        if isfile(lastversionfile):
            f = open(lastversionfile, 'r')
            LastVersionData = f.read()
            f.close()
            logger.debug("CurrentVersionData is %s." % CurrentVersionData)
            logger.debug("LastVersionData is %s." % LastVersionData)
            #compare with current data
            if (LastVersionData != CurrentVersionData):
                # another version has been installed and tested
                f = open(lastversionfile, 'w')
                f.write(CurrentVersionData+"\n")
                f.flush()
                f.close()
        else:
            f = open(lastversionfile,'w')
            f.write(CurrentVersionData+"\n")
            f.flush()
            f.close()

    def _HeartBeat(self):
        config = Ganga.Utility.Config.getConfig('Configuration')
        heartbeatfile = str(config['gangadir'])+os.sep+"heartbeat.txt"
        heartbeattmpfile = str(config['gangadir'])+os.sep+"heartbeattemp.txt"
        try:
            f = open(heartbeatfile,'r')
            HeartBeatData = (f.read())[:-7].strip()
            f.close()
            HBtime = HeartBeatData[:15].strip()
            HeartBeatTime = datetime.datetime(*(time.strptime(HBtime,"%H:%M:%S %j %y")[0:6]))
            if ((datetime.datetime.now() - HeartBeatTime) > datetime.timedelta(minutes=10)):
                f = open(heartbeattmpfile, 'w')
                f.write(datetime.datetime.now().strftime("%H:%M:%S %j %y")+" - "+str(os.getpid()))
                f.close()
                try:
                    shutil.move(heartbeattmpfile,heartbeatfile)
                except:
                    raise GangaRobotContinueError("Failed to move heartbeat file")            
        except IOError as e:
            f  = open(heartbeatfile, 'w')
            f.write(datetime.datetime.now().strftime("%H:%M:%S %j %y")+" - "+str(os.getpid()))
            f.close()
#




# Bootstrap all of ganga, setup GPI, registries, etc.
from Ganga.Utility.Runtime import allRuntimes
from Ganga.Utility.Config import getConfig
from Ganga.Utility.logging import getLogger
from Ganga.Utility.Plugin import allPlugins
from Ganga.GPIDev.Base import ProtectedAttributeError, ReadOnlyObjectError, GangaAttributeError
from Ganga.GPIDev.Lib.Job.Job import JobError
from Ganga import _gangaPythonPath
from Ganga.GPIDev.Base.Proxy import GPIProxyObjectFactory, getProxyClass
from Ganga.GPIDev.Credentials import getCredential
from Ganga.GPIDev.Persistency import export, load
from Ganga.GPIDev.Adapters.IPostProcessor import MultiPostProcessor
from Ganga.Runtime import Repository_runtime
import Ganga.Core
from Ganga.GPIDev.Lib.JobTree import TreeError
from Ganga.Core.GangaRepository import getRegistry
from Ganga.GPIDev.Base.VPrinter import full_print
from Ganga.GPIDev.Base.Proxy import implRef
import Ganga.GPIDev.Lib.Config
from Ganga.Utility.feedback_report import report
from Ganga.Runtime.gangadoc import adddoc
from Ganga.Core.GangaThread.WorkerThreads.ThreadPoolQueueMonitor import ThreadPoolQueueMonitor
from Ganga.Runtime import plugins

logger = getLogger(modulename=True)


def exportToPublicInterface(name, object, doc_section, docstring=None):
    """
    export the given functions/objects to both Ganga.GPI and ganga
    Note that GPI should be retired in time
    """
    import ganga
    import Ganga.GPI

    # export to GPI
    setattr(Ganga.GPI, name, object)
    setattr(ganga, name, object)
    adddoc(name, object, doc_section, docstring)


def ganga_license():
    'Print the full license (GPL)'
    with open(os.path.join(_gangaPythonPath, '..', 'LICENSE_GPL')) as printable:
        logger.info(printable.read())


def typename(obj):
    """Return a name of Ganga object as a string, example: typename(j.application) -> 'DaVinci'"""
    from Ganga.GPIDev.Base.Proxy import isProxy, stripProxy, implRef
    if isProxy(obj):
        if hasattr(stripProxy(obj), '_name'):
            return stripProxy(obj)._name
        else:
            logger = Ganga.Utility.logging.getLogger()
            logger.error("Object %s DOES NOT have the _name parameter set" % (str(obj)))
            return ""
    else:
        if hasattr(obj, '_name'):
            return obj._name
        else:
            logger = Ganga.Utility.logging.getLogger()
            logger.error("Object %s DOES NOT have the %s or _name parameter set" % (str(obj), str(implRef)))
            return ""


def categoryname(obj):
    """Return a category of Ganga object as a string, example: categoryname(j.application) -> 'applications'"""
    from Ganga.GPIDev.Base.Proxy import isProxy, stripProxy, implRef
    if isProxy(obj):
        if hasattr(stripProxy(obj), '_category'):
            return stripProxy(obj)._category
        else:
            logger = Ganga.Utility.logging.getLogger()
            logger.error("Object %s DOES NOT have the _category parameter set" % (str(obj)))
            return ""
    else:
        if hasattr(obj, '_category'):
            return obj._category
        else:
            logger = Ganga.Utility.logging.getLogger()
            logger.error("Object %s DOES NOT have the %s or _category parameter set" % (str(obj), str(implRef)))
            return ""


def plugins(category=None):
    """List loaded plugins.

    If no argument is given return a dictionary of all loaded plugins.
    Keys are category name. Values are lists of plugin names in each
    category.

    If a category is specified (for example 'splitters') return a list
    of all plugin names in this category.
    """
    from Ganga.Utility.Plugin import allPlugins
    if category:
        return allPlugins.allClasses(category).keys()
    else:
        d = {}
        for c in allPlugins.allCategories():
            d[c] = allPlugins.allCategories()[c].keys()
        return d

def convert_merger_to_postprocessor(j):
    from Ganga.GPIDev.Base.Proxy import stripProxy
    if len(stripProxy(j.postprocessors).process_objects):
        logger.info('job(%s) already has postprocessors' % j.fqid)
    if stripProxy(j).merger is None:
        logger.info(
            'job(%s) does not have a merger to convert' % j.fqid)
    if not len(stripProxy(j.postprocessors).process_objects) and stripProxy(j).merger is not None:
        mp = MultiPostProcessor()
        mp.process_objects.append(stripProxy(j).merger)
        stripProxy(j).postprocessors = mp


# ------------------------------------------------------------------------------------
# Setup the shutdown manager
from Ganga.Core.InternalServices import ShutdownManager
ShutdownManager.install()

# ------------------------------------------------------------------------------------
# start queues
exportToPublicInterface('queues', ThreadPoolQueueMonitor(), 'Objects')

# ------------------------------------------------------------------------------------
# Bootstrap all runtimes (e.g. GangaLHCb, GangaDirac, GangaAtlas, etc.)
for n, r in zip(allRuntimes.keys(), allRuntimes.values()):
    try:
        r.bootstrap(Ganga.GPI.__dict__)
    except Exception as err:
        logger.error('problems with bootstrapping %s -- ignored', n)
        logger.error('Reason: %s' % str(err))
        raise err
    try:
        r.loadNamedTemplates(Ganga.GPI.__dict__,
                             Ganga.Utility.Config.getConfig('Configuration')['namedTemplates_ext'],
                             Ganga.Utility.Config.getConfig('Configuration')['namedTemplates_pickle'])
    except Exception as err:
        logger.error('problems with loading Named Templates for %s', n)
        logger.error('Reason: %s' % str(err))

for r in allRuntimes.values():
    try:
        r.loadPlugins()
    except Exception as err:
        logger.error("problems with loading plugins for %s -- ignored" % r.name)
        logger.error('Reason: %s' % str(err))

# ------------------------------------------------------------------------------------
# make all plugins visible in GPI
for k in allPlugins.allCategories():
    for n in allPlugins.allClasses(k):
        cls = allPlugins.find(k, n)
        if not cls._declared_property('hidden'):
            exportToPublicInterface(n, getProxyClass(cls), 'Classes')

# ------------------------------------------------------------------------------------
# set the default value for the plugins
default_plugins_cfg = getConfig("Plugins")

for opt in default_plugins_cfg:
    try:
        category, tag = opt.split('_')
    except ValueError, err:
        logger.warning("do not understand option %s in [Plugins]", opt)
        logger.debug('Reason: want %s' % str(err))
    else:
        if tag == 'default':
            try:
                allPlugins.setDefault(category, default_plugins_cfg[opt])
            except Ganga.Utility.Plugin.PluginManagerError as x:
                logger.warning('cannot set the default plugin "%s": %s' % (opt, x))
        else:
            logger.warning("do not understand option %s in [Plugins]", opt)

# ------------------------------------------------------------------------------------
# set alias for default Batch plugin (it will not appear in the configuration)
# batch_default_name = getConfig('Configuration').getEffectiveOption('Batch')
# try:
#     batch_default = allPlugins.find('backends', batch_default_name)
# except Exception as x:
#     raise Ganga.Utility.Config.ConfigError('Check configuration. Unable to set default Batch backend alias (%s)' % str(x))
# else:
#     allPlugins.add(batch_default, 'backends', 'Batch')
#     exportToPublicInterface('Batch', batch_default._proxyClass, 'Classes')

# ------------------------------------------------------------------------------------
# Add errors to public interface
exportToPublicInterface('GangaAttributeError', GangaAttributeError, 'Exceptions')
exportToPublicInterface('ProtectedAttributeError', ProtectedAttributeError, 'Exceptions')
exportToPublicInterface('ReadOnlyObjectError', ReadOnlyObjectError, 'Exceptions')
exportToPublicInterface('JobError', JobError, 'Exceptions')

# ------------------------------------------------------------------------------------
# only the available credentials are exported
credential = getCredential(name='GridProxy', create=False)
if credential:
    exportToPublicInterface('gridProxy', GPIProxyObjectFactory(credential), 'Objects', 'Grid proxy management object.')

credential = getCredential('AfsToken')
if credential:
    exportToPublicInterface('afsToken', GPIProxyObjectFactory(credential), 'Objects', 'AFS token management object.')

# ------------------------------------------------------------------------------------
# Add Misc functions to public interface
exportToPublicInterface('license', ganga_license, 'Functions')
exportToPublicInterface('load', load, 'Functions')
exportToPublicInterface('export', export, 'Functions')
exportToPublicInterface('typename', typename, 'Functions')
exportToPublicInterface('categoryname', categoryname, 'Functions')
exportToPublicInterface('plugins', plugins, 'Functions')
exportToPublicInterface('convert_merger_to_postprocessor',
                       convert_merger_to_postprocessor, 'Functions')
exportToPublicInterface('config', Ganga.GPIDev.Lib.Config.config,
                       'Objects', 'access to Ganga configuration')
exportToPublicInterface('ConfigError', Ganga.GPIDev.Lib.Config.ConfigError,
                       'Exceptions')
exportToPublicInterface('report', report, 'Functions')

# ------------------------------------------------------------------------------------
# bootstrap the repositories and connect to them
for n, k, d in Repository_runtime.bootstrap():
    # make all repository proxies visible in GPI
    exportToPublicInterface(n, k, 'Objects', d)

# ------------------------------------------------------------------------------------
# JobTree
jobtree = GPIProxyObjectFactory(getRegistry("jobs").getJobTree())
exportToPublicInterface(
    'jobtree', jobtree, 'Objects', 'Logical tree view of the jobs')
exportToPublicInterface('TreeError', TreeError, 'Exceptions')

# ------------------------------------------------------------------------------------
# ShareRef
shareref = GPIProxyObjectFactory(getRegistry("prep").getShareRef())
exportToPublicInterface('shareref', shareref, 'Objects',
            'Mechanism for tracking use of shared directory resources')

# ------------------------------------------------------------------------------------
# export full_print
exportToPublicInterface('full_print', full_print, 'Functions')

# ------------------------------------------------------------------------------------
#  bootstrap core modules
interactive = False
Ganga.Core.bootstrap(getattr(Ganga.GPI.jobs, implRef), interactive)
Ganga.GPIDev.Lib.Config.bootstrap()

# ------------------------------------------------------------------------------------
# run post bootstrap hooks
for r in allRuntimes.values():
    try:
        r.postBootstrapHook()
    except Exception as err:
        logger.error("problems with post bootstrap hook for %s" % r.name)
        logger.error("Reason: %s" % str(err))



from GangaAtlas.Lib.ATLASDataset.DQ2Dataset import DQ2Dataset
from Ganga.GPIDev.Lib.Tasks.common import *
from Ganga.GPIDev.Lib.Tasks.ITransform import ITransform
from Ganga.GPIDev.Lib.Job.Job import JobError
from Ganga.GPIDev.Lib.Registry.JobRegistry import JobRegistrySlice, JobRegistrySliceProxy
from Ganga.Core.exceptions import ApplicationConfigurationError
from Ganga.GPIDev.Lib.Tasks.ITransform import ITransform
from Ganga.GPIDev.Lib.Tasks.TaskLocalCopy import TaskLocalCopy
from GangaAtlas.Lib.Tasks.AtlasUnit import AtlasUnit
from GangaAtlas.Lib.ATLASDataset.DQ2Dataset import DQ2Dataset, DQ2OutputDataset
from GangaAtlas.Lib.ATLASDataset.ATLASDataset import ATLASLocalDataset, ATLASOutputDataset
from GangaAtlas.Lib.Athena.DQ2JobSplitter import DQ2JobSplitter
from dq2.clientapi.DQ2 import DQ2, DQUnknownDatasetException, DQDatasetExistsException, DQFileExistsInDatasetException, DQInvalidRequestException
from dq2.container.exceptions import DQContainerAlreadyHasDataset, DQContainerDoesNotHaveDataset
from GangaAtlas.Lib.ATLASDataset.DQ2Dataset import dq2_lock, dq2
from dq2.common.DQException import DQException
from Ganga.GPIDev.Schema import *
from Ganga.GPIDev.Base.Proxy import addProxy, stripProxy
import Ganga.GPI as GPI
import os

from Ganga.Utility.Config import getConfig
configDQ2 = getConfig('DQ2')

class AtlasTransform(ITransform):
   _schema = Schema(Version(1,0), dict(ITransform._schema.datadict.items() + {
      'local_location'     : SimpleItem(defvalue='', doc='Local location to copy output to', typelist=["str"]),
      'include_file_mask'       : SimpleItem(defvalue = [], typelist=['str'], sequence=1, doc = 'List of Regular expressions of which files to include in copy'),
      'exclude_file_mask'       : SimpleItem(defvalue = [], typelist=['str'], sequence=1, doc = 'List of Regular expressions of which files to exclude from copy'),
      'files_per_job'     : SimpleItem(defvalue=-1, doc='files per job (cf DQ2JobSplitter.numfiles)', modelist=["int"]),
      'MB_per_job'     : SimpleItem(defvalue=0, doc='Split by total input filesize (cf DQ2JobSplitter.filesize)', modelist=["int"]),
      'subjobs_per_unit'     : SimpleItem(defvalue=0, doc='split into this many subjobs per unit master job (cf DQ2JobSplitter.numsubjobs)', modelist=["int"]),
      'rebroker_fraction'    : SimpleItem(defvalue=0.6, doc='Fraction of failed subjobs to complete subjobs above which the job will be rebrokered', modelist=["float"]),
      'num_dq2_threads'     : SimpleItem(defvalue=1, copyable=1, doc='Number of DQ2 download threads to run simultaneously (use setNumDQ2Threads to modify after submission)', typelist=["int"]),
      'files_per_unit' : SimpleItem(defvalue=-1, copyable=1, doc='Maximum number of files to assign to each unit from the given local files (i.e. AtlasLocalDataset). If < 1, use all files. At present, does not apply to DQ2Datasets', typelist=["int"]),
    }.items()))

   _category = 'transforms'
   _name = 'AtlasTransform'
   _exportmethods = ITransform._exportmethods + [ 'addUnit', 'getContainerName', 'initializeFromContainer', 'initializeFromDatasets', 'checkOutputContainers', 'setNumDQ2Threads', 'checkInputDatasets', 'changeDownloadLocation' ]

   def __init__(self):
      super(AtlasTransform,self).__init__()

      # force a delay of 1 minute to ensure DQ2 datasets have been registered properly
      self.chain_delay = 5

   def setNumDQ2Threads(self, num_threads):
      """Set the number of threads"""
      self.num_dq2_threads = num_threads
      
   def check(self):
      """Additional checks to base class"""

      # base class first
      super(AtlasTransform,self).check()
      
      # if a local copy has been specified, create the DSs required
      if self.local_location != '':
         self.unit_copy_output = TaskLocalCopy()
         self.unit_copy_output.local_location = self.local_location
         self.unit_copy_output.include_file_mask = self.include_file_mask
         self.unit_copy_output.exclude_file_mask = self.exclude_file_mask

   def checkOutputContainers(self):
      """Go through all completed units and make sure datasets are registered as required"""
      logger.info("Cleaning out transform %d container..." % self.getID())

      try:
         dslist = []
         dq2_lock.acquire()
         try:
            dslist = dq2.listDatasetsInContainer(self.getContainerName())
         except:
            dslist = []

         try:
            dq2.deleteDatasetsFromContainer(self.getContainerName(), dslist )

         except DQContainerDoesNotHaveDataset:
            pass
         except Exception as x:
            logger.error("Problem cleaning out Transform container: %s %s", x.__class__, x)
         except DQException as x:
            logger.error('DQ2 Problem cleaning out Transform container: %s %s' %( x.__class__, x))
      finally:
         dq2_lock.release()

      logger.info("Checking output data has been registered for Transform %d..." % self.getID())
      for unit in self.units:
         
         if len(unit.active_job_ids) == 0:
            continue

         if unit.status == "completed" and GPI.jobs(unit.active_job_ids[0]).outputdata and GPI.jobs(unit.active_job_ids[0]).outputdata._impl._name == "DQ2OutputDataset":
            logger.info("Checking containers in Unit %d..." % unit.getID() )
            unit.registerDataset()            

   def createUnits(self):
      """Create new units if required given the inputdata"""
      
      # call parent for chaining
      super(AtlasTransform,self).createUnits()
      
      # if there is no input data, just create a single unit given the application
      if len(self.inputdata) == 0 and len(self.units) == 0:
         unit = AtlasUnit()
         unit.name = "Unit %d" % len(self.units)
         self.addUnitToTRF( unit )
      
      # loop over input data and see if we need to create any more units
      for inds in self.inputdata:
         
         ok = True

         if inds._name == "DQ2Dataset":
            # check if this data is being run over
            ok = False
            for unit in self.units:
               if unit.inputdata.dataset == inds.dataset:
                  ok = True

            if not ok:
               # new unit required for this dataset
               unit = AtlasUnit()
               unit.name = "Unit %d" % len(self.units)
               self.addUnitToTRF( unit )
               unit.inputdata = inds

         elif inds._name == "ATLASLocalDataset":

            # different behaviour depending on files_per_unit
            if self.files_per_unit < 0:
               # check if this data is being run over
               ok = False
               for unit in self.units:
                  if set(unit.inputdata.names) == set(inds.names):
                     ok = True
                     
               if not ok:
                  # new unit required for this dataset
                  unit = AtlasUnit()
                  unit.name = "Unit %d" % len(self.units)
                  self.addUnitToTRF( unit )
                  unit.inputdata = inds
            
            else:

               ok = False
               curr_data = []
               for unit in self.units:
                  curr_data.extend( unit.inputdata.names )

               if set(inds.names) in set( curr_data ) or set(inds.names) == set( curr_data ):
                  ok = True

               if not ok:
                  # new unit(s) required for this dataset
                  num = 0
                  while num < len( inds.names ):
                     unit = AtlasUnit()
                     unit.name = "Unit %d" % len(self.units)
                     self.addUnitToTRF( unit )
                     unit.inputdata = inds.clone()
                     unit.inputdata.names = inds.names[num:num + self.files_per_unit]
                     num += self.files_per_unit


   def addUnit(self, outname, dsname, template = None):
      """Create a new unit based on this ds and output"""
      unit = AtlasUnit()
      if not template:
         unit.inputdata = DQ2Dataset()
      else:
         unit.inputdata = stripProxy( template )
      unit.inputdata.dataset = dsname
      unit.name = outname
      self.addUnitToTRF( unit )

   def getContainerName(self, max_length = configDQ2['OUTPUTDATASET_NAMELENGTH'] - 2):
      """Return the container for this transform"""
      if self.name == "":
         name = "trf"
      else:
         name = self.name
         
      dsn = [self._getParent().getContainerName()[:-1], name, "%i/" % self.getID() ] 
      if len(".".join(dsn)) > max_length:
         # too big so force a reduction of Task Name and compress trf name
         dsn2 = [self._getParent().getContainerName(max_length / 2)[:-1], "", "%i/" % self.getID() ]
         dsn =  [self._getParent().getContainerName(max_length / 2)[:-1],
                 name[: - (len(".".join(dsn2)) - max_length) ], "%i/" % self.getID() ]

      return (".".join(dsn)).replace(":", "_").replace(" ", "").replace(",","_")

   def createChainUnit( self, parent_units, use_copy_output = True ):
      """Create an output unit given this output data"""
      
      # we need valid parent jobs
      for parent in parent_units:
         # need datasetname filled for Panda jobs
         if len(parent.active_job_ids) == 0 or \
                (GPI.jobs(parent.active_job_ids[0]).application._impl._name != "TagPrepare" and \
                 GPI.jobs(parent.active_job_ids[0]).outputdata and \
                 GPI.jobs(parent.active_job_ids[0]).backend._impl._name == "Panda" and \
                 GPI.jobs(parent.active_job_ids[0]).outputdata.datasetname == ""):
            return None

         # need datasetList filled for Jedi jobs
         if len(parent.active_job_ids) == 0 or \
                (GPI.jobs(parent.active_job_ids[0]).application._impl._name != "TagPrepare" and \
                 GPI.jobs(parent.active_job_ids[0]).outputdata and \
                 GPI.jobs(parent.active_job_ids[0]).backend._impl._name == "Jedi" and \
                 len(GPI.jobs(parent.active_job_ids[0]).outputdata.datasetList) == 0):
            return None

         # for local jobs, make sure units are complete
         if GPI.jobs(parent_units[0].active_job_ids[0]).outputdata._impl._name == "ATLASOutputDataset" and \
                parent.status != "completed":
            return None
                

      # Are we doing Local -> Local? i.e. are we going from ATLASOutputDataset?
      # Problem: Doesn't take into account merger locations...
      if GPI.jobs(parent_units[0].active_job_ids[0]).outputdata._impl._name == "ATLASOutputDataset":
         unit = AtlasUnit()
         unit.inputdata = ATLASLocalDataset()

         for parent in parent_units:
            for l in GPI.jobs(parent.active_job_ids[0]).outputdata.output:
               unit.inputdata.names += l
            
      # should we use the copy_output (ie. local output). Special case for TagPrepare
      elif GPI.jobs(parent_units[0].active_job_ids[0]).application._impl._name == "TagPrepare":
         
         # make sure all have completed before taking the tag-info
         if parent_units[0].status != "completed":
            return None
         
         unit = AtlasUnit()
         unit.inputdata = DQ2Dataset()
         unit.inputdata.tag_info = GPI.jobs(parent_units[0].active_job_ids[0]).application.tag_info
         
      elif not use_copy_output or not parent.copy_output:
         unit = AtlasUnit()
         unit.inputdata = DQ2Dataset()
         ds_list = []
         for parent in parent_units:
            
            # Don't just use the main datasetname as Jedi introduces separate containers for logs and output files
            if GPI.jobs(parent.active_job_ids[0]).backend._impl._name == "Jedi":
               for ds in GPI.jobs(parent.active_job_ids[0]).outputdata.datasetList:
                  if not ds.endswith(".log/"):
                     unit.inputdata.dataset.append( ds )
            else:
               unit.inputdata.dataset.append( GPI.jobs(parent.active_job_ids[0]).outputdata.datasetname )
         
      else:

         unit = AtlasUnit()
         unit.inputdata = ATLASLocalDataset()

         for parent in parent_units:
            # unit needs to have completed and downloaded
            if parent.status != "completed":
               return None

            # we should be OK so copy all output to an ATLASLocalDataset
            for f in parent.copy_output.files:
               unit.inputdata.names.append( os.path.join( parent.copy_output.local_location, f ) )
         
      return unit
   
   def initializeFromContainer(self, dset, template = None, using_jedi = True):
      """Initialise the trf with given container, creating a unit for each DS"""
      if dset[-1] != "/":
         logger.error("Please supply a container!")
         return
      
      if not using_jedi:
         try:
            tid_datasets = dq2.listDatasetsInContainer(dset)
         except DQUnknownDatasetException:
            logger.error("dataset container %s not found" % dset)
            return
         
         logger.info("Found %i datasets matching %s..." % (len(tid_datasets), dset))
         
         for ds in tid_datasets:
            self.addUnit('.'.join( ds.split(".")[1:-1] ), ds, template)
      else:
         self.addUnit('.'.join( dset[:-1].split(".")[1:] ), dset, template)


   def initializeFromDatasets(self, dset_list, template = None, using_jedi = True):
      """Initialise the trf with the given dataset list, creating a unit for each DS"""

      for ds in dset_list:
         
         if ds[-1] == "/":
            if not using_jedi:
               try:
                  tid_datasets = dq2.listDatasetsInContainer(ds)
               except DQUnknownDatasetException:
                  logger.error("dataset container %s not found" % ds)
         
               logger.info("Found %i datasets matching %s..." % (len(tid_datasets), ds))
         
               for ds2 in tid_datasets:
                  self.addUnit('.'.join( ds.split(".")[1:-1] ), ds2, template)
            else:
               self.addUnit('.'.join( ds[:-1].split(".")[1:] ), ds, template)
         else:
            self.addUnit('.'.join( ds.split(".")[1:-1] ), ds, template)
            
   def checkInputDatasets(self):
      """Check the distribution of the input datasets"""
      if self.backend._name != "Panda" and self.backend._name != "Jedi":
         return

      for unit in self.units:
         print "Checking %s..." % ','.join( unit.inputdata.dataset )
         loc = unit.inputdata.get_locations()
         non_tape = []
         ok_sites = []
         for s in loc:
            if s.find("TAPE") != -1:
               continue
            non_tape.append(s)

            if s in self.backend.requirements.excluded_sites:
               continue               
            ok_sites.append(s)

         # non tape sites
         if len(non_tape) == 0:
            print "ERROR: No non-tape site available for DS '%s'" % unit.inputdata.dataset
         elif len(non_tape) == 1:
            print "WARNING: Only one non-tape site available for DS '%s'" % unit.inputdata.dataset

         # excluded sites
         if len(ok_sites) == 0:
            print "ERROR: No non-excluded sites available for DS '%s'" % unit.inputdata.dataset
         elif len(ok_sites) == 1:
            print "WARNING: Only one non-excluded site available for DS '%s'" % unit.inputdata.dataset

         # finally check panda brokerage
         from GangaPanda.Lib.Panda import selectPandaSite
         site = "NONE"
         try:
            site = selectPandaSite( self, ':'.join( ok_sites ) )
         except:
            pass

         if site == "NONE":
            print "ERROR: No non-blcklisted, non-Tape, non-excluded sites available for DS '%s'" % unit.inputdata.dataset
              
   def changeDownloadLocation(self, new_location, move_files = True):
      """Change the local download area"""
      import os

      # loop over the units
      self.local_location = new_location
      for unit in self.units:

         # do we have a valid local download?
         if not unit.copy_output or unit.copy_output._name != "TaskLocalCopy":
            logger.error("No locel copy available for unit %d" % unit.getID())
            continue
         
         if unit.copy_output.local_location == new_location:
            continue

         # grab the download lock
         if not unit._acquireDownloadLock():
            logger.error("Could not get download lock after 10s for unit %d - maybe try calling the function again?" % unit.getID())
            continue

         # move any files that have already been copied
         if move_files:
            import shutil
            os.makedirs(new_location)
            for fname in unit.copy_output.files:
               old_path = os.path.join(unit.copy_output.local_location, fname)
               new_path = os.path.join(new_location, fname)
               shutil.move(old_path, new_path)
         else:
            unit.copy_output.files = []

         # change the location
         unit.copy_output.local_location = new_location
         
         # set unit to running so the download is run again
         if unit.status == "completed":
            unit.updateStatus("running")
            
         if self.status == "completed":
            self.updateStatus("running")

         # release download lock
         unit._releaseDownloadLock()




from copy import deepcopy

from LHCbAnalysisTransform import LHCbAnalysisTransform
from LHCbAnalysisTransform import job_colours
from LHCbAnalysisTransform import markup
from LHCbAnalysisTransform import partition_colours
from Ganga.GPIDev.Lib.Tasks.Task import Task
from Ganga.GPIDev.Base.Proxy import stripProxy
from Ganga.GPIDev.Base.Proxy import isType
from Ganga.Core.exceptions import GangaException
from Ganga.Core.exceptions import GangaAttributeError
from Ganga.Utility.logging import getLogger
from GangaLHCb.Lib.LHCbDataset.BKQuery import BKQuery
logger = getLogger(modulename=True)

# to do...
# 1) Multi threading, even with multi threads, bottle neck at the server while updating.
# 2) Check prepared works(looks like does because of application.clone())
# 3) Dirac bulk submit
# 4) Sort out job resubmission on failure - done
# 5) Mergers - done
# 6) submit_counter for master job = 1 always unless whole job resubmitted. - done
# 7) Partition updating from completed by manual resubmit - done


class LHCbAnalysisTask(Task):

    """LHCbAnalysisTask - Top level class that looks after groups of LHCbAnalysisTransforms.

    The LHCbAnalysisTask class looks after the running of LHCb Analysis jobs, including helping to keep
    them up to date when new data is added or removed.

    Create a task using the following syntax

    In[1]: t = LHCbAnalysisTask()

    Append a transform to the task as below:

    In[2]: tr = LHCbAnalysisTransform()
    In[3]: t.appendTransform(tr)

    Or use an existing transform as a template to clone many new transforms, each with their
    own bookkeeping query (BKQuery) object.

    In[4]: t.addQuery(tr,[BKQuery('<dataset1>'), BKQuery(<dataset2>)])

    To set task running, do:

    In[5]: t.run()

    Can view the tasks registry by using either of the following

    In[6]: tasks          # less verbose
    In[7]: tasks.table()  # more verbose

    A Pretty, operational view of all the partition and subjobs within a task can be obtained with.

    In[8]: t.overview()

    Finally can update all the transforms within a task with.

    In[9]: t.update()"""

    _schema = Task._schema.inherit_copy()
    _schema.datadict['name'].defvalue = 'LHCbAnalysisTask'
    _category = 'tasks'
    _name = 'LHCbAnalysisTask'
    _exportmethods = Task._exportmethods
    _exportmethods += ['addQuery', 'update']

    default_registry = "tasks"

    # Public GPI methods
    #####################################################################

    #/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
    def addQuery(self, transform, bkQuery, associate=True):
        """Allows the user to add multiple transforms corresponding to the list of
        BKQuery type objects given in the second argument. The first argument
        is a transform object to use as the basis for the creation of further
        transforms."""
        if not isType(transform, LHCbAnalysisTransform):
            raise GangaException(
                None, 'First argument must be an LHCbAnalysisTransform objects to use as the basis for establishing the new transforms')

        # Check if the template transform is associated with the Task
        try:
            self.transforms.index(transform)
        except:
            if associate:
                logger.info(
                    'The transform is not associated with this Task, doing so now.')
                self.appendTransform(transform)

        # Check if the BKQuery input is correct and append/update
        if type(bkQuery) is not list:
            bkQuery = [bkQuery]
        for bk in bkQuery:
            if not isType(bk, BKQuery):
                raise GangaAttributeError(
                    None, 'LHCbTransform expects a BKQuery object or list of BKQuery objects passed to the addQuery method')
            if transform.query is None:  # If template has no query itself
                logger.info('Attaching query to transform')
                transform.query = stripProxy(bk)
                transform.update()
            else:  # Duplicate from template
                logger.info('Duplicating transform to add new query.')
                tr = deepcopy(transform)
                tr.query = stripProxy(bk)
                self.appendTransform(tr)

    #/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
    def appendTransform(self, transform):
        """Append a transform to this task. This method also performs an update() on
        the transform once successfully appended."""
        r = super(LHCbAnalysisTask, self).appendTransform(transform)
        if hasattr(transform, 'task_id'):
            stripProxy(transform).task_id = self.id
        else:
            raise GangaException(None, 'Couldnt set the task id')
        if hasattr(transform, 'transform_id'):
            try:
                stripProxy(transform).transform_id = self.transforms.index(transform)
            except:
                raise GangaException(
                    None, 'transform not added to task properly')
        else:
            raise GangaException(None, 'Coundnt set the transform id')

        # User may wish to append a transform and then update the whole task
        # after appending a query later.
        if transform.query is not None:
            transform.update()
        self.updateStatus()
        return r

    #/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
    def help(self):
        """Brief description of the LHCbAnalysisTask object."""
        logger.info(
            "This is an LHCbTask, Which simplifies the query driven analysis of data")

    #/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
    def overview(self):
        """ Get an ascii art overview over task status."""
        logger.info("Partition Colours: " + ", ".join([markup(key, partition_colours[key])
                                                       for key in ["hold", "ready", "running", "completed", "attempted", "failed", "bad", "unknown"]]))
        logger.info("Job Colours: " + ", ".join([markup(job, job_colours[job])
                                                 for job in ["new", "submitting", "submitted", "running", "completing", "completed", "failed", "killed", "incomplete", "unknown"]]))
        logger.info(
            "Lists the transforms, their partitions and partition subjobs, as well as the number of failures.")
        logger.info("Format: (partition/subjob number)[:(number of failures)]")
        logger.info('')
        for t in self.transforms:
            t.overview()

    #/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
    def update(self, resubmit=False):
        """Update the dataset information of all attached transforms. This will
        include any new data in the processing or re-run jobs that have data which
        has been removed."""
        # Tried to use multithreading, better to check the tasksregistry class
        # Also tried multiprocessing but bottleneck at server.
        for t in self.transforms:
            try:
                t.update(resubmit)
            except GangaException as e:
                logger.warning(e.__str__())
                continue

    # Public methods
    #####################################################################

    #/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
    # Called as part of the monitoring loop, which doesn't call this
    # while task in 'new'
    def updateStatus(self):
        """Check and update the status of each attached transform and then update
        own status."""
        if self.status is not 'new':  # only want to start this when user runs the task else jobs submitted when appending
            for t in self.transforms:  # could thread this
                t.checkStatus()

        # The overridden method will update the tasks status based on the
        # transforms status
        return super(LHCbAnalysisTask, self).updateStatus()

# End of class LHCbAnalysisTask
########################################################################



from Ganga.GPIDev.Lib.Tasks.common import *
from Ganga.GPIDev.Lib.Tasks.ITransform import ITransform
from Ganga.GPIDev.Lib.Job.Job import JobError
from Ganga.GPIDev.Lib.Registry.JobRegistry import JobRegistrySlice, JobRegistrySliceProxy
from Ganga.Core.exceptions import ApplicationConfigurationError
from Ganga.GPIDev.Lib.Tasks.ITransform import ITransform
from Ganga.GPIDev.Lib.Tasks.TaskLocalCopy import TaskLocalCopy

from GangaNA62.Lib.Tasks.NA62Unit import NA62Unit

import os
from commands import getstatusoutput

class NA62Transform(ITransform):
   _schema = Schema(Version(1,0), dict(ITransform._schema.datadict.items() + {
      'num_jobs' : SimpleItem(defvalue=-1,typelist=['int'],doc="Number of jobs with this application to process"),
    }.items()))

   _category = 'transforms'
   _name = 'NA62Transform'
   _exportmethods = ITransform._exportmethods + [ ]

   def __init__(self):
      super(NA62Transform,self).__init__()

   def update(self):
      "Catch the update to check the DB"
      nec_file = ".gpytho"
      work_dir = "/clusterhome/home/protopop"
      nec_str = open(os.path.join( work_dir, nec_file )).read().strip().strip('#')
      mysqlc = "mysql -hhughnon.ppe.gla.ac.uk -ugridbot -p%s -s GridJobs" % nec_str

      # check for pending actions
      rc, out = getstatusoutput("echo \"SELECT COUNT(*) FROM events WHERE attr2 LIKE 'c%%-action-2%%';\" | %s" % mysqlc)
      num_clone = int(out)
      rc, out = getstatusoutput("echo \"SELECT COUNT(*) FROM events WHERE attr2 LIKE 's%%-action-2%%';\" | %s" % mysqlc)
      num_submit = int(out)

      if (num_clone % 2 == 1) or (num_submit % 2 == 1):
         logger.warning( "num. Clone (%d), Num. Submit (%d). Pending requests. Waiting." % (num_clone, num_submit) )
         return 0

      # check for number of jobs
      status_cond = "status!='CLEARED' AND status!='CANCELLED' AND status!='FAILED' AND status!='ABORTED' AND status IS NOT NULL"
      rc, out = getstatusoutput("echo \"SELECT COUNT(run) FROM jobs WHERE %s;\" | %s" % (status_cond, mysqlc))
      num_jobs = int(out)
      
      if (num_jobs > 300):
         logger.warning( "Too many jobs in the system" )
         return 0
      
      return super(NA62Transform, self).update()

   def createUnits(self):
      """Create new units if required given the inputdata"""
      
      # call parent for chaining
      super(NA62Transform,self).createUnits()
      
      # Given the number of jobs, see if we should create more units
      if len(self.units) == 0:
         for i in range(0, self.num_jobs):            
            unit = NA62Unit()
            unit.name = "sub%d" % i
            self.addUnitToTRF( unit )




from __future__ import absolute_import

from .common import logger
from Ganga.Utility.ColourText import status_colours, overview_colours, ANSIMarkup
markup = ANSIMarkup()

from .TaskApplication import taskApp
from Ganga.GPIDev.Base import GangaObject
from Ganga.GPIDev.Base.Proxy import getName
from Ganga.GPIDev.Base.Proxy import stripProxy
from Ganga.GPIDev.Lib.Job.Job import JobError
from Ganga.GPIDev.Lib.Registry.JobRegistry import JobRegistrySlice, JobRegistrySliceProxy
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem, ComponentItem, FileItem
from Ganga import GPI


class Transform(GangaObject):
    _schema = Schema(Version(1, 0), {
        'status': SimpleItem(defvalue='new', protected=1, copyable=0, doc='Status - running, pause or completed', typelist=["str"]),
        'name': SimpleItem(defvalue='Simple Transform', doc='Name of the transform (cosmetic)', typelist=["str"]),
        'application': ComponentItem('applications', defvalue=None, optional=1, load_default=False, filter="checkTaskApplication", doc='Application of the Transform. Must be a Task-Supporting application.'),
        'inputsandbox': FileItem(defvalue=[], typelist=['str', 'Ganga.GPIDev.Lib.File.File.File'], sequence=1, doc="list of File objects shipped to the worker node "),
        'outputsandbox': SimpleItem(defvalue=[], typelist=['str'], sequence=1, doc="list of filenames or patterns shipped from the worker node"),
        'inputdata': ComponentItem('datasets', defvalue=None, optional=1, load_default=False, doc='Input dataset'),
        'outputdata': ComponentItem('datasets', defvalue=None, optional=1, load_default=False, doc='Output dataset'),
        'backend': ComponentItem('backends', defvalue=None, optional=1, load_default=False, doc='Backend of the Transform.'),
        'run_limit': SimpleItem(defvalue=4, doc='Number of times a partition is tried to be processed.', protected=1, typelist=["int"]),
        '_partition_status': SimpleItem(defvalue={}, hidden=1, copyable=0, doc='Map (only necessary) partitions to their status'),
        '_app_partition': SimpleItem(defvalue={}, hidden=1, copyable=0, doc='Map of applications to partitions'),
        '_app_status': SimpleItem(defvalue={}, hidden=1, copyable=0, doc='Map of applications to status'),
        '_next_app_id': SimpleItem(defvalue=0, hidden=1, copyable=0, doc='Next ID used for the application', typelist=["int"]),
    })

    _category = 'transforms'
    _name = 'Transform'
    _exportmethods = [
        'run', 'pause',  # Operations
        'setPartitionStatus', 'setRunlimit', 'setFailed',  # Control Partitions
        'getPartitionStatus', 'getJobs', 'getPartitionJobs',
        # Info
        'overview', 'info', 'n_all', 'n_status', 'retryFailed'
    ]

#   _app_status = {}
    _partition_apps = None

    # possible partition status values:
    # ignored, hold, ready, running, completed, attempted, failed, bad

# Special methods:
    def __init__(self):
        super(Transform, self).__init__()
        self.initialize()

    def _readonly(self):
        """A transform is read-only if the status is not new."""
        if self.status == "new":
            return 0
        return 1

    def initialize(self):
        from Ganga import GPI
        self.backend = stripProxy(GPI.Local())

    def check(self):
        pass

    def startup(self):
        """This function is used to set the status after restarting Ganga"""
        # Make sure that no partitions are kept "running" from previous
        # sessions
        clist = self._partition_status.keys()
        for c in clist:
            self.updatePartitionStatus(c)
        # At this point the applications still need to notify the Transformation of their status
        # Search jobs for task-supporting applications
        id = "%i:%i" % (
            self._getParent().id, self._getParent().transforms.index(self))
        for j in GPI.jobs:
            if "tasks_id" in stripProxy(j.application).getNodeData():
                # print "tasks_id of jobid ", j.fqid,
                # stripProxy(j.application).getNodeAttribute("tasks_id"), id
                if stripProxy(j.application).getNodeAttribute("tasks_id").endswith(id):
                    try:
                        if j.subjobs:
                            for sj in j.subjobs:
                                app = stripProxy(sj.application)
                                stripProxy(app.getTransform()).setAppStatus(
                                    app, app._getParent().status)
                        else:
                            app = stripProxy(j.application)
                            stripProxy(app.getTransform()).setAppStatus(
                                app, app._getParent().status)
                    except AttributeError as e:
                        logger.error("%s", e)

    def getPartitionApps(self):
        if self._partition_apps is None:
            # Create the reverse map _partition_apps from _app_partition
            self._partition_apps = {}
            for (app, partition) in self._app_partition.iteritems():
                if partition in self._partition_apps:
                    if not app in self._partition_apps[partition]:
                        self._partition_apps[partition].append(app)
                else:
                    self._partition_apps[partition] = [app]
        return self._partition_apps

    def fix(self):
        """This function fixes inconsistencies in application status"""
        # Create the reverse map _partition_apps from _app_partition
        self._app_status = {}
        # Make sure that no partitions are kept "running" from previous
        # sessions
        clist = self._partition_status.keys()
        for c in clist:
            self.updatePartitionStatus(c)
        # At this point the applications still need to notify the Transformation of their status
        # Search jobs for task-supporting applications

        id = "%i:%i" % (
            self._getParent().id, self._getParent().transforms.index(self))
        for j in GPI.jobs:
            if "tasks_id" in stripProxy(j.application).getNodeData():
                if stripProxy(j.application).getNodeAttribute("tasks_id") == id:
                    try:
                        if j.subjobs:
                            for sj in j.subjobs:
                                app = stripProxy(sj.application)
                                stripProxy(app.getTransform()).setAppStatus(
                                    app, app._getParent().status)
                        else:
                            app = stripProxy(j.application)
                            stripProxy(app.getTransform()).setAppStatus(
                                app, app._getParent().status)
                    except AttributeError as e:
                        logger.error("%s", e)


# Public methods
    def run(self, check=True):
        """Sets this transform to running status"""
        if self.status == "new" and check:
            self.check()
        if self.status != "completed":
            self.updateStatus("running")
            #self.status = "running"
            # Check if this transform has completed in the meantime
            is_complete = True
            for s in self._partition_status.values():
                if s != "completed" and s != "bad":
                    is_complete = False
                    break
            if is_complete:
                self.updateStatus("completed")
                #self.status = "completed"
            task = self._getParent()
            if task:
                task.updateStatus()
        else:
            logger.warning("Transform is already completed!")

    def pause(self):
        """Pause the task - the background thread will not submit new jobs from this task"""
        if self.status != "completed":
            self.updateStatus("pause")
            #self.status = "pause"
            task = self._getParent()
            if task:
                task.updateStatus()
        else:
            logger.debug("Transform is already completed!")

    def setRunlimit(self, newRL):
        """Set the number of times a job should be resubmitted before the transform is paused"""
        self.run_limit = newRL
        cs = self._partition_status.items()
        for (c, s) in cs:
            if s in ["attempted", "failed"]:
                failures = self.getPartitionFailures(c)
                if failures >= newRL:
                    self._partition_status[c] = "failed"
                else:
                    self._partition_status[c] = "attempted"
        logger.debug("Runlimit set to %i", newRL)

    def setPartitionStatus(self, partition, status):
        """ Set the Status of the given partition to "ready", "hold", "bad" or "completed".
            The status is then updated to the status indicated by the applications"""
        self.setPartitionsStatus([partition], status)

    def getJobs(self):
        """ Get the job slice of all jobs for this transform """
        return self.getPartitionJobs(None)

    def getPartitionJobs(self, partition):
        """ Get the job slice that processed the given partition. Iterates over the job list. """
        task = self._getParent()
        id = task.transforms.index(self)
        if partition is None:
            sname = "tasks(%i).transforms[%i].getJobs()" % (task.id, id)
        else:
            sname = "tasks(%i).transforms[%i].getPartitionJobs(%s)" % (
                task.id, id, partition)
        jobslice = JobRegistrySlice(sname)

        def addjob(j):
            if partition is None or self._app_partition[j.application.id] == partition:
                jobslice.objects[j.fqid] = stripProxy(j)

        for j in GPI.jobs:
            try:
                stid = j.application.tasks_id.split(":")
                if int(stid[-2]) == task.id and int(stid[-1]) == id:
                    if j.subjobs:
                        for sj in j.subjobs:
                            addjob(sj)
                    else:
                        addjob(j)
            except Exception as err:
                logger.debug("getPartitionJobs Exception:\n%s" % str(err))
                pass
        return JobRegistrySliceProxy(jobslice)

    def setFailed(self, partition):
        """ Tells Tasks that all Applications that have executed this partition have actually failed."""
        for aid in self._app_partition:
            if aid in self._app_status and self._app_status[aid] == "removed":
                continue
            # Save the status
            self._app_status[aid] = "failed"
            # Update the corresponding partition status
        self.setPartitionStatus(partition, "ready")

    def retryFailed(self):
        """Retry all failed partitions (forget about failed jobs)"""
        for aid in self._app_partition:
            if aid in self._app_status and self._app_status[aid] == "failed":
                self._app_status[aid] = "removed"
        clist = self._partition_status.keys()
        for c in clist:
            self.updatePartitionStatus(c)

# Internal methods
    def finalise(self):
        """Finalise the transform - no-op by default"""
        return

    def submitJobs(self, n):
        """Create Ganga Jobs for the next N partitions that are ready and submit them."""
        next = self.getNextPartitions(n)
        if len(next) == 0:
            return 0
        numjobs = 0
        for j in self.getJobsForPartitions(next):
            stripProxy(j.application).transition_update("submitting")
            try:
                j.submit()
            except JobError:
                logger.error(
                    "Error on job submission! The current transform will be paused until this problem is fixed.")
                logger.error(
                    "type tasks(%i).run() to continue after the problem has been fixed.", self._getParent().id)
                self.pause()
            numjobs += 1
        return numjobs

    def checkTaskApplication(self, app):
        """warns the user if the application is not compatible """
        if app is None:
            return None
        if not "tasks_id" in stripProxy(app).getNodeData():
            return taskApp(app)
        return app

    def setAppStatus(self, app, new_status):
        """Reports status changes in application jobs
           possible status values: 
           normal   : (new, submitting,) submitted, running, completing, completed
           failures : killed, failed
           transient: incomplete (->new), unknown, removed"""

        # Check if we know the occurring application...
        if app.id == -1:
            return
        if not app.id in self._app_partition:
            logger.warning(
                "%s was contacted by an unknown application %i.", self.fqn(), app.id)
            return
        # Silently ignore message if the application is already removed or
        # completed
        if app.id in self._app_status and self._app_status[app.id] in ["removed", "completed", "failed"]:
            return
        # Check the status
        if new_status == "completed" and not self.checkCompletedApp(app):
            logger.error(
                "%s app %i failed despite listed as completed!", self.fqn(), app.id)
            new_status = "failed"
        # Save the status
        self._app_status[app.id] = new_status
        # Update the corresponding partition status
        self.updatePartitionStatus(self._app_partition[app.id])

    def setMasterJobStatus(self, job, new_status):
        """hook for a master job status update"""
        return

    def updatePartitionStatus(self, partition):
        """ Calculate the correct status of the given partition. 
            "completed" and "bad" is never changed here
            "hold" is only changed to "completed" here. """
        # print "updatePartitionStatus ", partition, " transform ", self.id
        # If the partition has status, and is not in a fixed state, check it!

        if partition in self._partition_status and (not self._partition_status[partition] in ["bad", "completed"]):
            # if we have no applications, we are in "ready" state
            if not partition in self.getPartitionApps():
                if self._partition_status[partition] != "hold":
                    self._partition_status[partition] = "ready"
            else:
                status = [self._app_status[app] for app in self.getPartitionApps()[partition]
                          if app in self._app_status and not self._app_status[app] in ["removed", "killed"]]
                # Check if we have completed this partition
                if "completed" in status:
                    self._partition_status[partition] = "completed"
                # Check if we are not on hold
                elif self._partition_status[partition] != "hold":
                    # Check if we are running
                    running = False
                    for stat in ["completing", "running", "submitted", "submitting"]:
                        if stat in status:
                            self._partition_status[partition] = "running"
                            running = True
                            break
                    if not running:
                        # Check if we failed
                        #failures = len([stat for stat in status if stat in ["failed","new"]])
                        failures = self.getPartitionFailures(partition)

                        if failures >= self.run_limit:
                            self._partition_status[partition] = "failed"
                        elif failures > 0:
                            self._partition_status[partition] = "attempted"
                        else:
                            # Here we only have some "unknown" applications
                            # This could prove difficult when launching new applications. Care has to be taken
                            # to get the applications out of "unknown" stats as quickly as possible, to avoid double submissions.
                            #logger.warning("Partition with only unknown applications encountered. This is probably not a problem.")
                            self._partition_status[partition] = "ready"
        # Notify the next transform (if any) of the change in input status
        self.notifyNextTransform(partition)

        # Update the Tasks status if necessary
        task = self._getParent()
        if partition in self._partition_status and self._partition_status[partition] in ["completed", "bad"] and self.status == "running":
            for s in self._partition_status.values():
                if s != "completed" and s != "bad":
                    return
            #self.status = "completed"
            self.updateStatus("completed")
            if task:
                task.updateStatus()
        elif self.status == "completed":
            for s in self._partition_status.values():
                if s != "completed" and s != "bad":
                    self.updateStatus("running")
                    #self.status = "running"
                    if task:
                        task.updateStatus()
                    return

    def notifyNextTransform(self, partition):
        """ Notify any dependant transforms of the input update """
        task = self._getParent()
        if task and (task.transforms.index(self) + 1 < len(task.transforms)):
            task.transforms[
                task.transforms.index(self) + 1].updateInputStatus(self, partition)

    def setPartitionsStatus(self, partitions, status):
        """ Set the Status of the partitions to "ready", "hold", "bad" or "completed".
            The status is then updated to the status indicated by the applications
            "bad" and "completed" is never changed except to "ignored", "hold" is only changed to "completed". """
        if status == "ignored":
            [self._partition_status.pop(
                c) for c in partitions if c in self._partition_status]
        elif status in ["ready", "hold", "bad", "completed"]:
            for c in partitions:
                self._partition_status[c] = status
        else:
            logger.error(
                "setPartitionsStatus called with invalid status string %s", status)
        for c in partitions:
            self.updatePartitionStatus(c)

    def setPartitionsLimit(self, limitpartition):
        """ Set all partitions from and including limitpartition to ignored """
        partitions = [c for c in self._partition_status if c >= limitpartition]
        self.setPartitionsStatus(partitions, "ignored")

    def getPartitionStatus(self, partition):
        if partition in self._partition_status:
            return self._partition_status[partition]
        else:
            return "ignored"

    def getNextPartitions(self, n):
        """Returns the N next partitions to process"""
        partitionlist = sorted(
            c for c, v in self._partition_status.items() if v in ["ready", "attempted"])
        return partitionlist[:n]

    def getNewAppID(self, partition):
        """ Returns a new application ID and associates this ID with the partition given. """
        id = self._next_app_id
        self._app_partition[id] = partition
        if partition in self.getPartitionApps():
            self.getPartitionApps()[partition].append(id)
        else:
            self.getPartitionApps()[partition] = [id]
        self._next_app_id += 1
        return id

    def createNewJob(self, partition):
        """ Returns a new job initialized with the transforms application, backend and name """
        task = self._getParent(
        )  # this works because createNewJob is only called by a task
        id = task.transforms.index(self)
        j = GPI.Job()
        stripProxy(j).backend = self.backend.clone()
        stripProxy(j).application = self.application.clone()
        stripProxy(j).application.tasks_id = "%i:%i" % (task.id, id)
        stripProxy(j).application.id = self.getNewAppID(partition)
        j.inputdata = self.inputdata
        j.outputdata = self.outputdata
        j.inputsandbox = self.inputsandbox
        j.outputsandbox = self.outputsandbox
        j.name = "T%i:%i C%i" % (task.id, id, partition)
        return j

# Methods that can/should be overridden by derived classes
    def checkCompletedApp(self, app):
        """Can be overriden to improve application completeness checking"""
        return True

    def updateInputStatus(self, ltf, partition):
        """Is called my the last transform (ltf) if the partition 'partition' changes status"""
        # per default no dependencies exist
        pass

    def getJobsForPartitions(self, partitions):
        """This is only an example, this class should be overridden by derived classes"""
        return [self.createNewJob(p) for p in partitions]

# Information methods
    def fqn(self):
        task = self._getParent()
        if task:
            return "Task %i Transform %i" % (task.id, task.transforms.index(self))
        else:
            return "Unassigned Transform '%s'" % (self.name)

    def n_all(self):
        return len(self._partition_status)

    def n_status(self, status):
        return len([cs for cs in self._partition_status.values() if cs == status])

    def overview(self):
        """ Get an ascii art overview over task status. Can be overridden """
        task = self._getParent()
        if not task is None:
            id = str(task.transforms.index(self))
        else:
            id = "?"
        o = markup("#%s: %s '%s'\n" % (id, getName(self), self.name), status_colours[self.status])
        i = 0
        partitions = sorted(self._partition_status.keys())
        for c in partitions:
            s = self._partition_status[c]
            if c in self.getPartitionApps():
                failures = self.getPartitionFailures(c)
                o += markup("%i:%i " % (c, failures), overview_colours[s])
            else:
                o += markup("%i " % c, overview_colours[s])
            i += 1
            if i % 20 == 0:
                o += "\n"
        logger.info(o)

    def info(self):
        logger.info(markup("%s '%s'" % (getName(self), self.name), status_colours[self.status]))
        logger.info("* backend: %s" % getName(self.backend))
        logger.info("Application:")
        self.application.printTree()

    def getPartitionFailures(self, partition):
        """Return the number of failures for this partition"""
        return len([1 for app in self.getPartitionApps()[partition] if app in self._app_status and self._app_status[app] in ["new", "failed"]])

    def updateStatus(self, status):
        """Update the transform status"""
        self.status = status



from Ganga.GPIDev.Adapters.IApplication import IApplication
from Ganga.GPIDev.Adapters.IPrepareApp import IPrepareApp
from Ganga.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from Ganga.GPIDev.Schema import *

from Ganga.Utility.Config import getConfig

from Ganga.GPIDev.Lib.File import *
#from Ganga.GPIDev.Lib.File import File
#from Ganga.GPIDev.Lib.File import SharedDir
from Ganga.GPIDev.Lib.Registry.PrepRegistry import ShareRef
from Ganga.GPIDev.Base.Proxy import isType
from Ganga.Core import ApplicationConfigurationError

from Ganga.GPIDev.Lib.File.SandboxFile import SandboxFile
from commands import getstatusoutput

import os, shutil
from Ganga.Utility.files import expandfilename
shared_path = os.path.join(expandfilename(getConfig('Configuration')['gangadir']),'shared',getConfig('Configuration')['user'])

class NA62MC(IPrepareApp):
    """
    NA62MC application -- running arbitrary programs.
    """
    _schema = Schema(Version(2,0), {
        'run_number' : SimpleItem(defvalue=-1,typelist=['int'],doc="Run Number to pass to the scripts"),
        'seed'       : SimpleItem(defvalue=-1,typelist=['int'],doc="Random seed to pass to the scripts - overwritten by run_number if not given"),
        'num_events' : SimpleItem(defvalue=-1,typelist=['int'],doc="Number of events to generate"),
        'decay_type' : SimpleItem(defvalue=-1,typelist=['int'],doc="Decay type to generate"),
        'decay_name' : SimpleItem(defvalue="",typelist=['str'],doc="String of the Decay type"),
        'mc_version' : SimpleItem(defvalue=8,typelist=['int'],doc="MC version to use"),
        'revision'   : SimpleItem(defvalue=0,typelist=['int'],doc="MC revision to use"),
        'script_name': SimpleItem(defvalue="",typelist=['str'],doc="The name of the script to download and use on submission"),
        'radcor'     : SimpleItem(defvalue=False,typelist=['bool'],doc="Use radiatve corrections or not"),
        'file_prefix': SimpleItem(defvalue='pluto',typelist=['str'],doc="Prefix for output file"),
        'job_type'   : SimpleItem(defvalue='prod',typelist=['str'],doc="Job type (e.g. prod or test)"),
        'is_prepared': SharedItem(defvalue=None, strict_sequence=0, visitable=1, copyable=1, typelist=['type(None)', 'bool', 'str'],protected=0,doc='Location of shared resources. Presence of this attribute implies the application has been prepared.'),
        'hash': SimpleItem(defvalue=None, typelist=['type(None)', 'str'], hidden=1, doc='MD5 hash of the string representation of applications preparable attributes')
        } )
    _category = 'applications'
    _name = 'NA62MC'
    _exportmethods = ['prepare', 'unprepare']

    def __init__(self):
        super(NA62MC,self).__init__()

    def getMACFileName(self):
        "return the mac filename"
        return "Run%d_d%d.mac" % (self.run_number, self.decay_type)

    def getDecayString(self):
        "return the decay sring from the DB"
        return self.decay_name

    def getRevision(self):
        "return the revision from the DB"
        return self.revision

    def getNextRunNumber(self):
        "return the next available run number from the DB"
        nec_file = ".gpytho"
        work_dir = "/clusterhome/home/protopop"
        nec_str = open(os.path.join( work_dir, nec_file )).read().strip().strip('#')
        mysqlc = "mysql -hhughnon.ppe.gla.ac.uk -ugridbot -p%s -s GridJobs" % nec_str
        rc, out = getstatusoutput("echo \"SELECT run FROM jobs ORDER BY run DESC LIMIT 1;\" | %s" % mysqlc)
        if rc != 0:
            logger.error(out)
        run_num = int(out) + 1
            
        return run_num
    
    def unprepare(self, force=False):
        """
        Revert the application back to it's unprepared state.
        """
        logger.debug('Running unprepare in NA62MC app')
        if self.is_prepared is not None:
            self.decrementShareCounter(self.is_prepared.name)
            self.is_prepared = None
        self.hash = None

    def prepare(self,force=False):
        """
        A method to place the NA62MC application into a prepared state.

        The application wil have a Shared Directory object created for it. 
        If the application's 'exe' attribute references a File() object or
        is a string equivalent to the absolute path of a file, the file 
        will be copied into the Shared Directory.

        Otherwise, it is assumed that the 'exe' attribute is referencing a 
        file available in the user's path (as per the default "echo Hello World"
        example). In this case, a wrapper script which calls this same command 
        is created and placed into the Shared Directory.

        When the application is submitted for execution, it is the contents of the
        Shared Directory that are shipped to the execution backend. 

        The Shared Directory contents can be queried with 
        shareref.ls('directory_name')
        
        See help(shareref) for further information.
        """

        if (self.is_prepared is not None) and (force is not True):
            raise Exception('%s application has already been prepared. Use prepare(force=True) to prepare again.'%(self._name))

        # sanity check values
        if (self.run_number == -1):
            raise Exception('Run number has not been set')

        if (self.decay_type == -1):
            raise Exception('decay type has not been set')
        
        if (self.num_events < 1):
            raise Exception('no events set to generate')

        if (self.script_name == ""):
          raise Exception('please provide the script to run')

        #lets use the same criteria as the configure() method for checking file existence & sanity
        #this will bail us out of prepare if there's somthing odd with the job config - like the executable
        #file is unspecified, has a space or is a relative path
        self.configure(self)
        logger.info('Preparing %s application.'%(self._name))
        self.is_prepared = ShareDir()
        logger.info('Created shared directory: %s'%(self.is_prepared.name))

        # download and store the WN script to run
        scr_path = os.path.join(shared_path,self.is_prepared.name, self.script_name)
        os.system("wget -q http://na62.gla.ac.uk/scripts/jw1v8.sh.txt -O %s" % scr_path)
        if not os.path.exists(scr_path):
            raise Exception('Could not download worker node script')

        # download and store the input files
        inp_path = os.path.join(shared_path,self.is_prepared.name, "input_files.tgz")
        os.system("wget -q http://na62.gla.ac.uk/input/input_files.tgz -O %s" % inp_path)
        if not os.path.exists(inp_path):
            raise Exception('Could not download the input files tarball')            

        # create the mac file and add to the shared area
        mac_out = ""
        for ln in open( os.path.join( os.path.dirname(__file__), "MAC_template.txt")).readlines():
            
            if ln.find("/output/fileName") != -1:
                # replace filename            
                mac_out += ("/output/fileName %s_v%d_r%d.root\n" % (self.file_prefix, self.mc_version, self.run_number))
            elif ln.find("/decay/type") != -1:
                # replace decay type            
                mac_out += ("/decay/type %d\n" % (self.decay_type))
            elif ln.find("/decay/radcor") != -1:
                # replace rad cor
                if (self.radcor):
                    mac_out += ("/decay/radcor 1\n")
                else:
                    mac_out += ("/decay/radcor 0\n")
            elif ln.find("/random/seedDecay") != -1:
                # replace seed decay
                seed = self.seed
                if (seed == -1):
                    seed = self.run_number
                    
                mac_out += ("/random/seedDecay %d\n" % (seed))
            elif ln.find("/run/number") != -1:
                # replace run number
                mac_out += ("/run/number %d\n" % (self.run_number))
            elif ln.find("/run/beamOn") != -1:
                # replace number of events
                mac_out += ("/run/beamOn %d\n" % (self.num_events))                
            else:
                mac_out += ln

        open(os.path.join(shared_path,self.is_prepared.name, self.getMACFileName()), "w").write(mac_out)

        #copy any 'preparable' objects into the shared directory
        send_to_sharedir = self.copyPreparables()
        #add the newly created shared directory into the metadata system if the app is associated with a persisted object
        self.checkPreparedHasParent(self)
        #return [os.path.join(self.is_prepared.name,os.path.basename(send_to_sharedir))]
        self.post_prepare()
        return 1


    def configure(self,masterappconfig):
        from Ganga.Core import ApplicationConfigurationError
        import os.path
        
        return (None,None)


class RTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from Ganga.GPIDev.Adapters.StandardJobConfig import StandardJobConfig

        #prepared_exe = File(os.path.join(shared_path,self.is_prepared.name, self.script_name))
        prepared_exe = File(os.path.join( os.path.dirname(__file__), "runNA62MC.sh"))
        #if app.is_prepared is not None:
        #    logger.info("Submitting a prepared application; taking any input files from %s" %(app.is_prepared.name))
        #    prepared_exe = File(os.path.join(os.path.join(shared_path,app.is_prepared.name),os.path.basename(prepared_exe.name)))

        inputbox = [ File(os.path.join(os.path.join(shared_path,app.is_prepared.name), app.getMACFileName())) ]
        inputbox.append( File(os.path.join(shared_path,app.is_prepared.name, app.script_name)) )
        inputbox.append( File(os.path.join(shared_path,app.is_prepared.name, "input_files.tgz")) )
        inputbox += app._getParent().inputsandbox

        #outputbox = [ 'na62run%d.stderr.err' % app.run_number, 'na62run%d.stdout.out' % app.run_number ]
        outputbox = []
        outputbox += app._getParent().outputsandbox

        env = {'NA62SCRIPT':app.script_name,
               'NA62STDOUT':'na62run%d.out' % app.run_number,
               'NA62STDERR':'na62run%d.err' % app.run_number}
        
        args = [app.getMACFileName(), "%s/r%d/%s" % (app.job_type, app.getRevision(), app.getDecayString())]
        c = StandardJobConfig(prepared_exe,inputbox,args,outputbox,env)
        return c
        

class LCGRTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from Ganga.Lib.LCG import LCGJobConfig

        prepared_exe = File(os.path.join( os.path.dirname(__file__), "runNA62MC.sh"))
        
        #if app.is_prepared is not None:
        #    logger.info("Submitting a prepared application; taking any input files from %s" %(app.is_prepared.name))
        #    prepared_exe = File(os.path.join(os.path.join(shared_path,app.is_prepared.name),os.path.basename(prepared_exe.name)))

        #outputbox = [ 'na62run%d.stderr.err' % app.run_number, 'na62run%d.stdout.out' % app.run_number ]
        outputbox = []
        outputbox += app._getParent().outputsandbox
        
        inputbox = [ File(os.path.join(os.path.join(shared_path,app.is_prepared.name), app.getMACFileName())) ]
        inputbox.append( File(os.path.join(shared_path,app.is_prepared.name, app.script_name)) )
        inputbox.append( File(os.path.join(shared_path,app.is_prepared.name, "input_files.tgz")) )
        inputbox += app._getParent().inputsandbox
        
        env = {'NA62SCRIPT':app.script_name,
               'NA62STDOUT':'na62run%d.out' % app.run_number,
               'NA62STDERR':'na62run%d.err' % app.run_number}
        args = [app.getMACFileName(), "%s/r%d/%s" % (app.job_type, app.getRevision(), app.getDecayString())]

        # add the output files
        app._getParent().outputfiles = [SandboxFile('na62run%d.out' % app.run_number), SandboxFile('na62run%d.err' % app.run_number),
                                        SandboxFile('__jdlfile__'), SandboxFile(app.getMACFileName()) ]
        
        return LCGJobConfig(prepared_exe,inputbox,args,outputbox,env)

from Ganga.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers

allHandlers.add('NA62MC','LSF', RTHandler)
allHandlers.add('NA62MC','Local', RTHandler)
allHandlers.add('NA62MC','PBS', RTHandler)
allHandlers.add('NA62MC','SGE', RTHandler)
allHandlers.add('NA62MC','Condor', RTHandler)
allHandlers.add('NA62MC','LCG', LCGRTHandler)
allHandlers.add('NA62MC','TestSubmitter', RTHandler)
allHandlers.add('NA62MC','Interactive', RTHandler)
allHandlers.add('NA62MC','Batch', RTHandler)
allHandlers.add('NA62MC','Cronus', RTHandler)
allHandlers.add('NA62MC','Remote', LCGRTHandler)
allHandlers.add('NA62MC','CREAM', LCGRTHandler)




from Ganga.GPIDev.Lib.Registry.TransientRegistry import TransientRegistry
from Ganga.Core.GangaRepository import addRegistry


def establishNamedTemplates(registryname, filebase, doc='', file_ext='tpl', pickle_files=False):
    j = TransientRegistry(
        registryname, filebase, doc, file_ext=file_ext, pickle_files=pickle_files)
    addRegistry(j)



from __future__ import absolute_import
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem, ComponentItem
from fnmatch import fnmatch
from Ganga.GPIDev.Adapters.IGangaFile import IGangaFile
from Ganga.Utility.logging import getLogger, _get_logging
from Ganga.GPIDev.Base.Proxy import isType, GPIProxyObjectFactory
from Ganga.Utility.Config import getConfig
import re
import copy
import glob
import os
import pickle
import stat
import Ganga.Utility.Config

logger = getLogger()
regex = re.compile('[*?\[\]]')
badlogger = getLogger('oauth2client.util')
badlogger.setLevel(_get_logging().ERROR)


class GoogleFile(IGangaFile):

    """
    The GoogleFile outputfile type allows for files to be directly uploaded, downloaded, removed and restored from the GoogleDrive service.
    It can be used as part of a job to output data directly to GoogleDrive, or standalone through the Ganga interface.

    example job: j=Job(application=Executable(exe=File('/home/hep/hs4011/Tests/testjob.sh'), args=[]),outputfiles=[GoogleFile('TestJob.txt')])

                 j.submit()

                 ### This job will automatically upload the outputfile 'TestJob.txt' to GoogleDrive.

    example of standalone submission:

                 g=GoogleFile('TestFile.txt')

                 g.localDir = '~/TestDirectory'        ### The file's location must be specified for standalone submission

                 g.put()                               ### The put() method uploads the file to GoogleDrive directly

    The GoogleFile outputfile is also compatible with the Dirac backend, making outputfiles from Dirac-run jobs upload directly to GoogleDrive.
    """

    _schema = Schema(Version(1, 1),
                     {'namePattern': SimpleItem(defvalue="", doc='pattern of the file name'),
                      'localDir': SimpleItem(defvalue="", copyable=1,
                                             doc='local dir where the file is stored, used from get and put methods'),
                      'subfiles': ComponentItem(category='gangafiles', defvalue=[], hidden=1,
                                                typelist=['Ganga.GPIDev.Lib.File.LCGSEFile'], sequence=1, copyable=0,
                                                doc="collected files from the wildcard namePattern"),
                      'failureReason': SimpleItem(defvalue="", copyable=1,
                                                  doc='reason for the upload failure'),
                      'compressed': SimpleItem(defvalue=False, typelist=['bool'], protected=0,
                                               doc='wheather the output file should be compressed before sending somewhere'),
                      'downloadURL': SimpleItem(defvalue="", copyable=1, protected=1,
                                                doc='download URL assigned to the file upon upload to GoogleDrive'),
                      'id': SimpleItem(defvalue="", copyable=1, hidden=1, protected=1,
                                       doc='GoogleFile ID assigned to file  on upload to GoogleDrive'),
                      'title': SimpleItem(defvalue="", copyable=1, hidden=1, protected=1,
                                                   doc='GoogleFile title of the uploaded file'),
                      'GangaFolderId': SimpleItem(defvalue="", copyable=1, hidden=1, protected=1,
                                                  doc='GoogleDrive Ganga folder  ID')
                      })
    _category = 'gangafiles'
    _name = 'GoogleFile'
    _exportmethods = ["get", "put", "remove", "restore", "deleteCredentials"]

    def __init__(self, namePattern=''):
        super(GoogleFile, self).__init__()
        self.namePattern = namePattern
        self.__initialized = False
        
        self.cred_path = os.path.join(getConfig('Configuration')['gangadir'], 'googlecreddata.pkl')

    def __initializeCred(self):
        while os.path.isfile(self.cred_path) == False:
            from oauth2client.client import OAuth2WebServerFlow

            # Copy your credentials from the APIs Console
#            CLIENT_ID = "54459939297.apps.googleusercontent.com"
#            CLIENT_SECRET = "mAToHx5RpXtwkeYR6nOIe_Yw"
            CLIENT_ID = '776655306197-dirtoquqsm7cpqgepvamofg5t2b5f637.apps.googleusercontent.com'
            CLIENT_SECRET = 'GpdEP-OBZZQLB3k-xxOpzFQG'
            # Check https://developers.google.com/drive/scopes for all
            # available scopes
            OAUTH_SCOPE = 'https://www.googleapis.com/auth/drive.file'

            # Redirect URI for installed apps
            REDIRECT_URI = 'urn:ietf:wg:oauth:2.0:oob'

            # Run through the OAuth flow and retrieve credentials
            credentials = ''
            flow = OAuth2WebServerFlow(
                CLIENT_ID, CLIENT_SECRET, OAUTH_SCOPE, REDIRECT_URI)
            authorize_url = flow.step1_get_authorize_url()
            try:
                import webbrowser
                webbrowser.get('macosx').open(authorize_url, 0, True)
            except:
                try:
                    import webbrowser
                    webbrowser.get(
                        'windows-default').open(authorize_url, 0, True)
                except:
                    try:
                        import webbrowser
                        webbrowser.get('firefox').open(authorize_url, 0, True)
                    except Exception, err:
                        logger.error("Error: %s" % str(err))
                        pass
            logger.info(
                'Go to the following link in your browser: ' + authorize_url)
            code = raw_input('Enter verification code: ').strip()
            try:
                credentials = flow.step2_exchange(code)
            except:
                deny = raw_input(
                    'An incorrect code was entered. Have you denied Ganga access to your GoogleDrive (y/[n])?')
                if deny == '' or deny[0:1].upper() == 'N':
                    pass
                elif deny[0:1].upper() == 'Y':
                    return None

            # Pickle credential data
            if credentials is not '':
                with open(self.cred_path, "wb") as output:
                    pickle.dump(credentials, output)
                
                os.chmod(self.cred_path, stat.S_IWUSR | stat.S_IRUSR)
                logger.info('Your GoogleDrive credentials have been stored in the file %s and are only readable by you. '
                            'The file will give permission to modify files in your GoogleDrive. '
                            'Permission can be revoked by going to "Manage Apps" in your GoogleDrive '
                            'or by deleting the credentials through the deleteCredentials GoogleFile method.' % self.cred_path)

        self.__initialized = True

        self._check_Ganga_folder()

    def __construct__(self, args):
        if (len(args) != 1) or (not isType(args[0], str)):
            super(GoogleFile, self).__construct__(args)
        else:
            self.namePattern = args[0]
        self.cred_path = os.path.join(getConfig('Configuration')['gangadir'], 'googlecreddata.pkl')

    def _attribute_filter__set__(self, n, v):
        if n == 'localDir':
            return os.path.expanduser(os.path.expandvars(v))
        return v

    def setLocation(self):
        """
        Sets the location of output files that were uploaded from the WN
        """
        raise NotImplementedError

    def location(self):
        """
        Return list with the locations of the post processed files (if they were configured to upload the output somewhere)
        """
        raise NotImplementedError

    def _on_attribute__set__(self, obj_type, attrib_name):
        r = copy.deepcopy(self)
        from Ganga.GPIDev.Lib.Job import Job
        if isinstance(obj_type, Job) and attrib_name == 'outputfiles':
            r.localDir = None
            r.failureReason = ''
        return r

    def deleteCredentials(self):
        """
        Deletes the user's GoogleDrive credentials

            example use: GoogleFile().deleteCredentials()
        """
        if self.__initilized == True:
            if os.path.isfile(self.cred_path) == True:
                os.remove(self.cred_path)
                logger.info('GoogleDrive credentials deleted')
                return None
        else:
            logger.info('There are no credentials to delete')

    def get(self):
        """
        Retrieves files uploaded to GoogleDrive through a job or by a standalone GoogleFile
        """
        service = self._setup_service()

        # Checks for wildcards and loops through get procedure for each result,
        # saving file to assigned directory
        if regex.search(self.namePattern) is not None:
            for f in self.subfiles:
                if f.downloadURL:
                    resp, content = service._http.request(f.downloadURL)
                    if resp.status == 200:
                        # print 'Status: %s' % resp
                        logger.info(
                            "File \'%s\' downloaded succesfully" % f.title)
                        dir_path = f.localDir
                        if f.localDir == '':
                            dir_path = self.localDir
                            if self.localDir == '':
                                dir_path = os.getcwd()
                        completeName = os.path.join(dir_path, f.title)
                        with open(completeName, "wb") as gotfile:
                            gotfile.write(content)

                    else:
                        # print 'An error occurred: %s' % resp
                        logger.info(
                            "Download unsuccessful, file \'%s\' may not exist on GoogleDrive" % f.title)
                else:
                    # The file doesn't have any content stored on Drive.
                    logger.info(
                        "No file \'%s\' exists on GoogleDrive" % f.title)
                    return None

        # Non-wildcard get request procedure
        else:
            if self.downloadURL:
                resp, content = service._http.request(self.downloadURL)
                if resp.status == 200:
                    # print 'Status: %s' % resp
                    logger.info("Download successful")
                    dir_path = self.localDir
                    if self.localDir == ('' or None):
                        dir_path = os.getcwd()
                    if self._getParent() is not None:
                        dir_path = self.getJobObject(
                        ).getOutputWorkspace().getPath()
                    completeName = os.path.join(dir_path, self.namePattern)
                    with open(completeName, "wb") as gotfile:
                        gotfile.write(content)
                else:
                    # print 'An error occurred: %s' % resp
                    logger.info(
                        "Download unsuccessful, the file may not exist on GoogleDrive")
                    return None
            else:
                # The file doesn't have any content stored on Drive.
                logger.info("No such file on GoogleDrive")
                return

    def getWNScriptDownloadCommand(self, indent):
        """
        Gets the command used to download already uploaded file
        """
        raise NotImplementedError

    def processWildcardMatches(self):
        raise NotImplementedError

    def __repr__(self):
        """
        Get the representation of the file
        """
        return "GoogleFile(namePattern='%s', downloadURL='%s')" % (self.namePattern, self.downloadURL)

    def put(self):
        """
        Postprocesses (upload) output file to the desired destination from the client
        """
        import hashlib
        from apiclient.http import MediaFileUpload

        service = self._setup_service()

        # Sets the target directory
        dir_path = self.localDir
        if self.localDir == '':
            dir_path = os.getcwd()

        if self._getParent() is not None:
            dir_path = self.getJobObject().getOutputWorkspace().getPath()

        # Wildcard procedure
        if regex.search(self.namePattern) is not None:
            for wildfile in glob.glob(os.path.join(dir_path, self.namePattern)):
                FILENAME = wildfile
                filename = os.path.basename(wildfile)

                # Upload procedure
                media_body = MediaFileUpload(
                    FILENAME, mimetype='text/plain', resumable=True)
                body = {
                    'title': '%s' % filename,
                    'description': 'A test document',
                    'mimeType': 'text/plain',
                    'parents': [{
                        "kind": "drive#fileLink",
                        "id": "%s" % self.GangaFolderId
                    }]
                }

                # Metadata file and md5checksum intergrity check
                file = service.files().insert(
                    body=body, media_body=media_body).execute()
                with open(FILENAME, 'rb') as thefile:
                    if file.get('md5Checksum') == hashlib.md5(thefile.read()).hexdigest():
                        logger.info("File \'%s\' uploaded successfully" % filename)
                    else:
                        logger.error(
                            "File \'%s\' uploaded unsuccessfully" % filename)

                # Assign new schema components to each file and append to job
                # subfiles
                g = GoogleFile(filename)
                g.downloadURL = file.get('downloadUrl', '')
                g.id = file.get('id', '')
                g.title = file.get('title', '')
                self.subfiles.append(GPIProxyObjectFactory(g))

        # For non-wildcard upload
        else:
            # Path to the file to upload
            FILENAME = os.path.join(dir_path, self.namePattern)

            # Upload procedure, can edit more of file metadata
            media_body = MediaFileUpload(
                FILENAME, mimetype='text/plain', resumable=True)
            body = {
                'title': '%s' % self.namePattern,
                'description': 'A test document',
                'mimeType': 'text/plain',
                'parents': [{
                    "kind": "drive#fileLink",
                    "id": "%s" % self.GangaFolderId
                }]
            }

            # Metadata storage and md5checksum integrity check
            file = service.files().insert(
                body=body, media_body=media_body).execute()

            with open(FILENAME, 'rb') as thefile:
                if file.get('md5Checksum') == hashlib.md5(thefile.read()).hexdigest():
                    logger.info("File \'%s\' uploaded succesfully" %
                                self.namePattern)
                else:
                    logger.error("Upload Unsuccessful")

            # Assign values to new schema components
            self.downloadURL = file.get('downloadUrl', '')
            self.id = file.get('id', '')
            self.title = file.get('title', '')

            return
        return GPIProxyObjectFactory(self.subfiles[:])

    def remove(self, permanent=False):
        """
        Move a file to the trash or permanently delete the file

            example use: GoogleFile().remove()

            or:          j = Job([...], outputfiles=GoogleFile()) --> j.submit --> j.outputfiles[0].remove()

        Remove multiple files by using

                         for i in j.outputfiles:
                             i.remove()

        The file can also be permanently deleted by using

                         GoogleFile().remove(True)

        However, this will make the file unrestorable
        """
        service = self._setup_service()

        from apiclient import errors

        # Wildcard procedure
        if regex.search(self.namePattern) is not None:
            for f in self.subfiles:
                if permanent == True:
                    try:
                        service.files().delete(fileId=f.id).execute()
                        f.downloadURL = ''
                        logger.info(
                            'File \'%s\' permanently deleted from GoogleDrive' % f.title)
                    except errors.HttpError as error:
                        # print 'An error occurred: %s' % error
                        logger.info(
                            'File \'%s\' deletion failed, or file already deleted' % f.title)
                else:
                    try:
                        service.files().trash(fileId=f.id).execute()
                        logger.info(
                            'File \'%s\' removed from GoogleDrive' % f.title)
                    except errors.HttpError as error:
                        # print 'An error occurred: %s' % error
                        logger.info(
                            'File \'%s\' removal failed, or file already removed' % f.title)

        # Non-wildcard request
        else:
            if permanent == True:
                try:
                    service.files().delete(fileId=self.id).execute()
                    self.downloadURL = ''
                    logger.info('File permanently deleted from GoogleDrive')
                except errors.HttpError as error:
                    # print 'An error occurred: %s' % error
                    logger.info(
                        'File deletion failed, or file already deleted')
            else:
                try:
                    service.files().trash(fileId=self.id).execute()
                    logger.info('File removed from GoogleDrive')
                except errors.HttpError as error:
                    # print 'An error occurred: %s' % error
                    logger.info('File removal failed, or file already removed')
                return None

    def restore(self):
        """
        Restore a file from the trash. This method will not work on permanently deleted files

            example use: GoogleFile().restore()
        """
        from apiclient import errors

        service = self._setup_service()

        # Wildcard procedure
        if regex.search(self.namePattern) is not None:
            for f in self.subfiles:
                try:
                    service.files().untrash(fileId=f.id).execute()
                    logger.info(
                        'File \'%s\' restored to GoogleDrive' % f.title)
                except errors.HttpError as error:
                    # print 'An error occurred: %s' % error
                    logger.info(
                        'File \'%s\' restore failed, or file does not exist on GoogleDrive' % f.title)

        # Non-wildcard request
        else:
            try:
                service.files().untrash(fileId=self.id).execute()
                logger.info('File restored to GoogleDrive')
            except errors.HttpError as error:
                # print 'An error occurred: %s' % error
                logger.info(
                    'File restore failed, or file does not exist on GoogleDrive')
            return None

    def _check_Ganga_folder(self):
        """
        Creates a Ganga folder on GoogleDrive if one is not already present
        """
        from apiclient import errors
        service = self._setup_service()

        page_token = None
        try:
            param = {}
            if page_token:
                param['pageToken'] = page_token
            files = service.files().list(**param).execute()
            items = files['items']
            for i in items:
                if i['title'] == 'Ganga':
                    self.GangaFolderId = i['id']
                    return
            page_token = files.get('nextPageToken')
        except errors.HttpError as error:
            logger.info('Failed to create Ganga folder on GoogleDrive')
            # print 'An error occurred: %s' % error

        body = {
            'title': 'Ganga',
            'description': 'A test folder',
            'mimeType': 'application/vnd.google-apps.folder'
        }
        file = service.files().insert(body=body).execute()
        self.GangaFolderId = file.get('id')

    def _setup_service(self):
        """
        Sets up the GoogleDrive service for other methods
        """
        from apiclient.discovery import build
        import httplib2
        http = httplib2.Http()
        if self.__initialized == False:
            self.__initializeCred()
        with open(self.cred_path, "rb") as nput:
            credentials = pickle.load(nput)
        http = credentials.authorize(http)
        service = build('drive', 'v2', http=http)
        return service

    def getWNInjectedScript(self, outputFiles, indent, patternsToZip, postProcessLocationsFP):
        """
        Returns script that have to be injected in the jobscript for postprocessing on the WN
        """
        logger.info('injecting')

    def _readonly(self):
        return False

    def _list_get__match__(self, to_match):
        if isinstance(to_match, str):
            return fnmatch(self.namePattern, to_match)
        if isinstance(to_match, type):
            # note stripProxy wont work on class types that aren't instances
            return isinstance(self, to_match._impl)
        return to_match == self

Ganga.Utility.Config.config_scope['GoogleFile'] = GoogleFile



from GangaTest.Framework.tests import GangaGPITestCase
from GangaTest.Framework.utils import sleep_until_completed
from Ganga.Utility.logging import getLogger
logger = getLogger()
#/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/
try:
    from Ganga.GPI import templatesLHCb, Job
except ImportError:
    templatesLHCb=[]
    pass

class TestTemplatesLHCb(GangaGPITestCase):

    def _check(self, template):
        logger.info("------------------------------------------------")
        logger.info("-    Now checking template: '%s'" % template.name)
        logger.info("------------------------------------------------")
        j = Job(template)
        j.submit()
        self.assertTrue(sleep_until_completed(j))

#/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/

dynamic_test_case = '''
def test_###NAME###(self):
    self._check(templatesLHCb('###NAME###'))
setattr(TestTemplatesLHCb, 'test_###NAME###', test_###NAME###)
'''
for t in templatesLHCb:
    exec(dynamic_test_case.replace('###NAME###', t.name))




from Ganga.GPIDev.Lib.Tasks.TaskApplication import taskify
from GangaAtlas.Lib.AthenaMC.AthenaMC import AthenaMC, AthenaMCSplitterJob
from GangaAtlas.Lib.Athena.Athena import Athena
AthenaTask = taskify(Athena,"AthenaTask")
AthenaMCTask = taskify(AthenaMC,"AthenaMCTask")
AthenaMCTaskSplitterJob = taskify(AthenaMCSplitterJob,"AthenaMCTaskSplitterJob")
from Ganga.GPIDev.Base.Proxy import addProxy, stripProxy

from Ganga.GPIDev.Schema import *
from Ganga.GPIDev.Adapters.ISplitter import ISplitter
class AnaTaskSplitterJob(ISplitter):
    """AnaTask handler for job splitting"""
    _name = "AnaTaskSplitterJob"
    _category = "splitters"
    _schema = Schema(Version(1,0), {
        'subjobs'           : SimpleItem(defvalue=[],sequence=1, doc="List of subjobs", typelist=["int"]),
        'numevtsperjob'     : SimpleItem(defvalue=0, doc='Number of events per subjob'),
        'numevtsperfile'    : SimpleItem(defvalue=0,doc='Maximum number of events in a file of input dataset')
    } )
    def split(self,job):
        from Ganga.GPIDev.Lib.Job import Job
        logger.debug("AnaTaskSplitterJob split called")
        sjl = []
        transform = stripProxy(job.application.getTransform())
        transform.setAppStatus(job.application, "removed")
        # Do the splitting
        for sj in self.subjobs:
            j = Job()
            j.inputdata = transform.partitions_data[sj-1]
            j.outputdata = job.outputdata
            j.application = job.application
            j.application.atlas_environment.append("OUTPUT_FILE_NUMBER=%i" % sj)
            j.backend = job.backend
            if transform.partitions_sites:
                if hasattr(j.backend.requirements, 'sites'):                
                    j.backend.requirements.sites = transform.partitions_sites[sj-1]                    
                else:
                    j.backend.site = transform.partitions_sites[sj-1]

            j.inputsandbox = job.inputsandbox
            j.outputsandbox = job.outputsandbox
            sjl.append(j)
            # Task handling
            j.application.tasks_id = job.application.tasks_id
            j.application.id = transform.getNewAppID(sj)
             #transform.setAppStatus(j.application, "submitting")
        if not job.application.tasks_id.startswith("00"):
            job.application.tasks_id = "00:%s" % job.application.tasks_id
        return sjl
    
from Ganga.GPIDev.Lib.Tasks.TaskApplication import task_map
task_map["Athena"] = AthenaTask
task_map["AthenaMC"] = AthenaMCTask



##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: StandardJobConfig.py,v 1.1 2008-07-17 16:40:52 moscicki Exp $
##########################################################################

from Ganga.Utility.logging import getLogger
import os
import copy
logger = getLogger()


class StandardJobConfig(object):

    """
    StandardJobConfig defines a standard input for many of the handlers: LSF, Localhost,LCG.
    It corresponds to a simplified JDL definition: specification of executable, arguments and input sandbox.
    Executable and arguments may be specified either as strings or File objects. In the second case they are
    automatically added to the input sandbox list.

    If you modify attributes of the StandardJobConfig object after the initialization, the you should
    do processValues() which perfomes a validation of the attributes and updates internal cache which is
    used by getter methods.

    """

    def __init__(self, exe=None, inputbox=None, args=None, outputbox=None, env=None):
        """
        exe - executable string to be run on the worker node or a File object to be shipped as executable script to the worker node
        args - list of strings which are passed as arguments to the executable string or File objects which are automatically added to the sandbox
        inputbox - list of additional File or FileBuffer objects which go to the sandbox
        outputbox - list of additional files which should be returned by the sandbox

        The constructor does processValues() automatically so the construction of the object may failed with exceptions raised by that method.
        Notes for derived classes:
          - this constructor should be called at the end of the derived constructor.
          - you may freely add new attributes as long as you they do not start with _
        """
        if exe is None:
            exe = ''
        if inputbox is None:
            inputbox = []
        if args is None:
            args = []
        if outputbox is None:
            outputbox = []
        self.exe = exe
        self.inputbox = inputbox[:]
        self.args = args
        self.outputbox = outputbox[:]
        self.env = env
        self.__all_inputbox = []
        self.__args_strings = []
        self.__exe_string = ""
        self.__sandbox_check = {}

        self.processValues()

    def __str__(self):
        config_str = "Config:"
        config_str += " Exe str: '%s'" % str(self.getExeString())
        config_str += " Arg: '%s'" % str(self.getArgStrings())
        config_str += " Input Sandbox files: '%s'" % str(self.getSandboxFiles())
        config_str += " Output files: '%s'" % str(self.getOutputSandboxFiles())
        return config_str

    def getSandboxFiles(self):
        '''Get all input sandbox files'''
        return self.__all_inputbox

    def getOutputSandboxFiles(self):
        """Get all output sandbox files. The duplicates are removed. """
        from Ganga.Utility.util import unique
        return unique(self.outputbox)

    def getExeString(self):
        '''Get a string which should be used at the worker node to invoke an executable.
        Note that this string does not necesserily have to be a file name on the worker node'''
        return self.__exe_string

    def getArgStrings(self):
        '''Get a list of strings which correspond to the arguments to the executable on the worker node.'''
        return self.__args_strings

    def processValues(self):
        '''Process original exe,args and inputbox values and extract strings suitable for the further processing.
        If the exe property is a File then this method will check if it has executable attributes.
        You do not have to call this method unless you explicitly modify some of the original values.
        '''

        self.__all_inputbox = self.inputbox[:]
        self.__args_strings = []
        self.__exe_string = ""
        self.__sandbox_check = {}

        def _get_path_in_sandbox(f):
            ''' A helper which checks conflicts in sandbox.
            If you try to add twice a file object f with the same name, it shows a warning.
            '''

            fn = f.getPathInSandbox()
            if fn in self.__sandbox_check:
                logger.warning('File %s already in the sandbox (source=%s). Overriding from source=%s', fn, self.__sandbox_check[fn], f.name)
            self.__sandbox_check[fn] = f.name
            return fn

        # to check for double file
        for f in self.inputbox:
            fn = _get_path_in_sandbox(f)

        # convert all args into strings (some of args may be FileItems)
        # make an assumption that all File Items go to the sandbox (thus
        # convert to basename)
        for a in self.args:
            if isinstance(a, str):
                self.__args_strings.append(a)
            else:
                try:
                    fn = _get_path_in_sandbox(a)
                    self.__args_strings.append(fn)
                    self.__all_inputbox.append(a)
                except AttributeError as x:
                    s = "cannot process argument %s, it is neither File nor string" % repr(
                        a)
                    logger.error(s)
                    raise ValueError(s)

        if isinstance(self.exe, str):
            self.__exe_string = self.exe
        else:
            try:
                self.__exe_string = _get_path_in_sandbox(self.exe)
                if not self.exe.isExecutable():
                    logger.warning(
                        'file %s is not executable, overriding executable permissions in the input sandbox' % self.exe.name)
                    self.exe.executable = True
                self.__all_inputbox.append(self.exe)
            except AttributeError as x:
                s = "cannot process exe property %s, it is neither File nor string (%s)" % (
                    repr(self.exe), str(x))
                logger.error(s)
                raise ValueError(s)

#
#
# $Log: not supported by cvs2svn $
# Revision 1.6.4.1  2007/10/30 15:19:47  moscicki
# obsoleted jobConfig.getExeCmdString() method
#
# Revision 1.6  2007/08/24 15:55:02  moscicki
# added executable flag to the file, ganga will set the executable mode of the app.exe file (in the sandbox only, the original file is not touched), this is to solve feature request #24452
#
# Revision 1.5  2006/08/07 12:09:06  moscicki
# bug #18271 bug fix from V.Romanovski
#
# Revision 1.4  2006/02/10 14:19:14  moscicki
# added outputsandbox
#
# Revision 1.3  2005/09/02 12:42:56  liko
# Extend StandardJobConfig with outputbox and environment
#
# Revision 1.2  2005/08/24 08:16:49  moscicki
# minor changes
#
#
#



##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: gangadoc.py,v 1.1 2008-07-17 16:41:00 moscicki Exp $
##########################################################################

"""Improved support for interactive help based on pydoc.
Remove hardcoded limits on the help text.
This module makes changes to the content of the pydoc module.
"""
from __future__ import absolute_import

import pydoc

_GPIhelp = '''Ganga Public Interface (GPI) Index

%(Classes)s

%(Exceptions)s

%(Functions)s

%(Objects)s

'''

from Ganga.Utility.logging import getLogger
logger = getLogger(modulename=True)

_GPIhelp_sections = {
    'Classes': [], 'Exceptions': [], 'Functions': [], 'Objects': []}


def adddoc(name, object, doc_section, docstring):
    '''
    Add automatic documentation to gangadoc system.
    "doc_section" specifies how the object should be documented.
    If docstring is specified then use it to document the object. Otherwise use __doc__ (via pydoc utilities).
    '''
    from Ganga.Utility.logic import implies
    assert(implies(docstring, doc_section == "Objects"))
    #assert(not docstring and not object.__doc__)

    _GPIhelp_sections[doc_section] += [(name, object, docstring)]


def makedocindex():
    '''
    Return a string with GPI Index.
    '''

    import pydoc

    sections = {}

    from Ganga.Utility.strings import ItemizedTextParagraph

    for sec, names in zip(_GPIhelp_sections.keys(), _GPIhelp_sections.values()):
        itbuf = ItemizedTextParagraph(sec + ':')
        for name, obj, docstring in names:
            # if docstring not provided when exporting the object to GPI then
            # use the docstring generated by pydoc
            if not docstring:
                docstring = pydoc.splitdoc(pydoc.getdoc(obj))[0]
            itbuf.addLine(name, docstring)

        sections[sec] = itbuf.getString()

    return _GPIhelp % sections


# modifications to pydoc module

# a workaround for hardcoded limit of maxlen==70 characters
class TextDoc2(pydoc.TextDoc):

    def docother(self, object, name=None, mod=None, maxlen=None, doc=None):
        return pydoc.TextDoc.docother(self, object, name=name, mod=mod, doc=doc)


class Helper2(pydoc.Helper):

    def __init__(self, *args, **kwds):
        pydoc.Helper.__init__(self, *args, **kwds)

    def intro(self):

        from Ganga.Runtime import _prog

        self.output.write("""************************************
%s

This is an interactive help based on standard pydoc help.

Type 'index'  to see GPI help index.
Type 'python' to see standard python help screen.
Type 'interactive' to get online interactive help from an expert.
Type 'quit'   to return to Ganga.
************************************
""" % (_prog.hello_string))

    def help(self, request):
        if isinstance(request, str):
            if request == 'python':
                pydoc.Helper.intro(self)
                return

            if request == 'index':
                self.output.write(makedocindex())
                return

            if request == 'interactive':
                from . import eliza
                self.output.write("""
This is an interactive help. At the prompt type your questions in plain english\n\n""")
                eliza.command_interface()
                return

            # eval the expression (in the context of Ganga.GPI namespace)
            from .gangadoceval import evaluate
            request = evaluate(request)

        return pydoc.Helper.help(self, request)

pydoc.text = TextDoc2()
pydoc.text._repr_instance.maxstring = 255
pydoc.text._repr_instance.maxother = 70

import sys

pydoc.help = Helper2(sys.stdin, sys.stdout)


# bugfix: #18012 overview: No help text avaialable for "jobs"
# with python2.4 should not be needed (?)
# patch the  doc method  from pydoc module  to support  documenting, in
# case of  af an instance,  the structure of  its class instead  of its
# value
# &
# bugfix: #12584
# overview: for some (unknown) reason the IPython code inserts an instance of
# IPython.FakeModule.FakeModule class in sys.modules
# inspect.getmodule(object) method raises in this a TypeError
# workaround: use an improved getmodule(object) method that does supplementary checks
# Notes:
# Running with Python2.2 IPython seems to insert a *stripped* version of FakeModule in sys.module which masks this bug
# This bug persists in IPython/Python 2.4 too


modulesbyfile = {}
import inspect


def mygetmodule(object):
    """Return the module an object was defined in, or None if not found."""
    import os.path
    if inspect.ismodule(object):
        return object
    if hasattr(object, '__module__'):
        return sys.modules.get(object.__module__)
    try:
        file = inspect.getabsfile(object)
    except TypeError:
        return None
    if file in modulesbyfile:
        return sys.modules.get(modulesbyfile[file])
    for module in sys.modules.values():
        # check if value is indeed a module
        if inspect.ismodule(module) and hasattr(module, '__file__'):
            modulesbyfile[
                os.path.realpath(inspect.getabsfile(module))] = module.__name__
    if file in modulesbyfile:
        return sys.modules.get(modulesbyfile[file])
    main = sys.modules['__main__']
    if not hasattr(object, '__name__'):
        return None
    if hasattr(main, object.__name__):
        mainobject = getattr(main, object.__name__)
        if mainobject is object:
            return main
    builtin = sys.modules['__builtin__']
    if hasattr(builtin, object.__name__):
        builtinobject = getattr(builtin, object.__name__)
        if builtinobject is object:
            return builtin


def doc2(thing, title='Python Library Documentation: %s', forceload=0):
    """Display text documentation, given an object or a path to an object."""
    try:
        object, name = pydoc.resolve(thing, forceload)
        desc = pydoc.describe(object)
        module = mygetmodule(object)
        if name and '.' in name:
            desc += ' in ' + name[:name.rfind('.')]
        elif module and module is not object:
            desc += ' in module ' + module.__name__

        if not (inspect.ismodule(object) or
                inspect.isclass(object) or
                inspect.isroutine(object) or
                isinstance(object, property)):
            # If the passed object is a piece of data or an instance,
            # document its available methods instead of its value.

            object = type(object)
            desc += ' object'
        pydoc.pager(title % desc + '\n\n' + pydoc.text.document(object, name))
    except (ImportError, pydoc.ErrorDuringImport) as value:
        logger.error(value)

pydoc.doc = doc2


#
#
# $Log: not supported by cvs2svn $
# Revision 1.6  2007/07/10 13:08:32  moscicki
# docstring updates (ganga devdays)
#
# Revision 1.5  2006/08/09 08:55:13  adim
# bugfix: #12584
# overview: IPython code inserts an instance of IPython.FakeModule.FakeModule class in sys.modules
# which breaks the help system.
#
# Revision 1.4  2006/07/12 12:40:45  moscicki
# bugfix: #18012 overview: No help text avaialable for "jobs"
#
# Revision 1.3  2006/02/10 14:17:08  moscicki
# fixed bugs:
# #14136 help system doesn't find documentation.
#
# Revision 1.2  2005/08/31 15:06:14  andrew
# add the interactive help system
#
# Revision 1.1  2005/08/24 15:24:11  moscicki
# added docstrings for GPI objects and an interactive ganga help system based on pydoc
#
#
#



import os
import os.path
import re
import datetime
from os.path import exists, isdir, realpath, isfile, islink
from os import pathsep, listdir, environ, fdopen
import subprocess
import tempfile
import Ganga.Utility.logging
import Ganga.Utility.Config
from optparse import OptionParser, OptionValueError

from Ganga.Utility.Config.Config import _after_bootstrap
from Ganga.Utility.logging import getLogger
logger = getLogger()

if not _after_bootstrap:
    configLHCb = Ganga.Utility.Config.makeConfig('LHCb', 'Parameters for LHCb')
    configDirac = Ganga.Utility.Config.getConfig('DIRAC')

    # Set default values for the LHCb config section.
    dscrpt = 'The name of the local site to be used for resolving LFNs into PFNs.'
    configLHCb.addOption('LocalSite', '', dscrpt)

    dscrpt = 'Files from these services will go to the output sandbox (unless \
    overridden by the user in a specific job via the Job.outputdata field). Files \
    from all other known handlers will go to output data (unless overridden by \
    the user in a specific job via the Job.outputsandbox field).'
    configLHCb.addOption('outputsandbox_types',
                     ['CounterSummarySvc', 'NTupleSvc',
                      'HistogramPersistencySvc', 'MicroDSTStream',
                      'EvtTupleSvc'], dscrpt)
    dscrpt = 'The string that is added after the filename in the options to tell' \
             ' Gaudi how to read the data. This is the default value used if the '\
             'file name does not match any of the patterns in '\
             'datatype_string_patterns.'
    configLHCb.addOption('datatype_string_default',
                     """TYP='POOL_ROOTTREE' OPT='READ'""", dscrpt)
    dscrpt = 'If a file matches one of these patterns, then the string here '\
         'overrides the datatype_string_default value.'
    defval = {"SVC='LHCb::MDFSelector'": ['*.raw', '*.RAW', '*.mdf', '*.MDF']}
    configLHCb.addOption('datatype_string_patterns', defval, dscrpt)
    configLHCb.addOption('UserAddedApplications', "", "List of user added LHCb applications split by ':'")

    configLHCb.addOption('SplitByFilesBackend', 'OfflineGangaDiracSplitter',
                     'Possible SplitByFiles backend algorithms to use to split jobs into subjobs,\
                      options are: GangaDiracSplitter, OfflineGangaDiracSplitter, splitInputDataBySize and splitInputData')

def _guess_version(name):
    if 'GANGASYSROOT' in os.environ.keys():
        gangasys = os.environ['GANGASYSROOT']
    else:
        raise OptionValueError("Can't guess %s version if GANGASYSROOT is not defined" % name)
    tmp = tempfile.NamedTemporaryFile(suffix='.txt')
    cmd = 'cd %s && cmt show projects > %s' % (gangasys, tmp.name)
    rc = subprocess.Popen([cmd], shell=True).wait()
    if rc != 0:
        msg = "Fail to get list of projects that Ganga depends on"
        raise OptionValueError(msg)
    p = re.compile(r'^\s*%s\s+%s_(\S+)\s+' % (name, name))
    for line in tmp:
        m = p.match(line)
        if m:
            version = m.group(1)
            return version
    msg = 'Failed to identify %s version that Ganga depends on' % name
    raise OptionValueError(msg)


def _store_root_version():
    if 'ROOTSYS' in os.environ:
        vstart = os.environ['ROOTSYS'].find('ROOT/') + 5
        vend = os.environ['ROOTSYS'][vstart:].find('/')
        rootversion = os.environ['ROOTSYS'][vstart:vstart + vend]
        os.environ['ROOTVERSION'] = rootversion
    else:
        msg = 'Tried to setup ROOTVERSION environment variable but no ROOTSYS variable found.'
        raise OptionValueError(msg)


def _store_dirac_environment():
    diracversion = _guess_version('LHCBDIRAC')
    import tempfile
    import subprocess
    import os
    platform = os.environ['CMTOPT']
    setup_script = 'SetupProject.sh'
    env = {}
    fdir = os.path.join(os.path.expanduser("~/.cache/LHCbDIRAC_ENV"), platform)
    if not os.path.exists(fdir):
        os.makedirs(fdir)
    fname = os.path.join(fdir, diracversion)
    if not os.path.exists(fname) or not os.path.getsize(fname):
        from Ganga.Utility.logging import getLogger
        log = getLogger()
        log.info("Storing new LHCbDirac environment (%s:%s)" % (str(diracversion), str(platform)))
        s_file = open(fname, 'w+')
        cmd = '/usr/bin/env bash -c \"source %s LHCBDIRAC %s ROOT>& /dev/null && '\
            'printenv > %s\"' % (setup_script, diracversion, fname)
        rc = subprocess.Popen([cmd], shell=True).wait()
        if rc != 0 or not os.path.exists(fname):
            msg = '--dirac: Failed to setup Dirac version %s as obtained from project dependency.' % diracversion
            raise OptionValueError(msg)
        count = 0
        for line in s_file.readlines():
            if line.find('DIRAC') >= 0:
                count += 1
            varval = line.strip().split('=')
            env[varval[0]] = ''.join(varval[1:])
        s_file.close()
        if count == 0:
            msg = 'Tried to setup Dirac version %s. For some reason this did not setup the DIRAC environment.' % diracversion
            raise OptionValueError(msg)
    os.environ['GANGADIRACENVIRONMENT'] = fname

if not _after_bootstrap:
    _store_dirac_environment()
    configDirac.setSessionValue('DiracEnvFile', os.environ['GANGADIRACENVIRONMENT'])

    _store_root_version()


def getEnvironment(config=None):
    import sys
    import os.path
    import PACKAGE

    PACKAGE.standardSetup()
    return {}


def loadPlugins(config=None):
    logger.debug("Importing Backends")
    import Lib.Backends
    logger.debug("Importing Applications")
    import Lib.Applications
    logger.debug("Importing LHCbDataset")
    import Lib.LHCbDataset
    logger.debug("Importing Mergers")
    import Lib.Mergers
    logger.debug("Importing RTHandlers")
    import Lib.RTHandlers
    logger.debug("Importing Splitters")
    import Lib.Splitters
    logger.debug("Importing Tasks")
    import Lib.Tasks
    logger.debug("Importing Files")
    import Lib.Files
    logger.debug("Importing Checkers")
    import Lib.Checkers
    logger.debug("Importing LHCbTasks")
    import Lib.Tasks
    logger.debug("Finished Importing")

from Ganga.GPIDev.Lib.File.Configure import outputconfig

outputconfig.overrideDefaultValue('FailJobIfNoOutputMatched', 'False')


# This is being dropped from 6.1.0 due to causing some bug in loading large numbers of jobs
#
# This will be nice to re-add once there is lazy loading support passed to the display for the 'jobs' command 09/2015 rcurrie
#
#from Ganga.GPIDev.Lib.Registry.JobRegistry import config as display_config
#display_config.overrideDefaultValue( 'jobs_columns', ('fqid', 'status', 'name', 'subjobs', 'application', 'backend', 'backend.actualCE', 'backend.extraInfo', 'comment') )
#display_config.overrideDefaultValue( 'jobs_columns_functions', {'comment': 'lambda j: j.comment', 'backend.extraInfo': 'lambda j : j.backend.extraInfo ', 'subjobs': 'lambda j: len(j.subjobs)', 'backend.actualCE': 'lambda j:j.backend.actualCE', 'application': 'lambda j: j.application._name', 'backend': 'lambda j:j.backend._name'} )
#display_config.overrideDefaultValue('jobs_columns_width', {'fqid': 8, 'status': 10, 'name': 10, 'application': 15, 'backend.extraInfo': 30, 'subjobs': 8, 'backend.actualCE': 17, 'comment': 20, 'backend': 15} )




import commands, exceptions, random, re, sys, time

from Ganga.Core.exceptions import ApplicationConfigurationError
from Ganga.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from Ganga.Core import BackendError

from GangaAtlas.Lib.ATLASDataset.DQ2Dataset import getDatasets

import Ganga.Utility.logging
logger = Ganga.Utility.logging.getLogger()

from Ganga.Utility.Config import getConfig
configPanda = getConfig('Panda')

def getLatestDBReleaseCaching():
    import tempfile
    import cPickle as pickle
    from pandatools import Client
    from GangaAtlas.Lib.Credentials.ProxyHelper import getNickname

    TMPDIR = tempfile.gettempdir()
    nickname = getNickname(allowMissingNickname=False)
    DBRELCACHE = '%s/ganga.latestdbrel.%s'%(TMPDIR,nickname)

    try:
        fh = open(DBRELCACHE)
        dbrelCache = pickle.load(fh)
        fh.close()
        if dbrelCache['mtime'] > time.time() - 3600:
            logger.debug('Loading LATEST DBRelease from local cache')
            return dbrelCache['atlas_dbrelease']
        else:
            raise Exception()
    except:
        logger.debug('Updating local LATEST DBRelease cache')
        atlas_dbrelease = Client.getLatestDBRelease(False)
        dbrelCache = {}
        dbrelCache['mtime'] = time.time()
        dbrelCache['atlas_dbrelease'] = atlas_dbrelease
        fh = open(DBRELCACHE,'w')
        pickle.dump(dbrelCache,fh)
        fh.close()
        return atlas_dbrelease


class ProdTransPandaRTHandler(IRuntimeHandler):
    """Runtime handler for the ProdTrans application."""

    def master_prepare(self, app, appmasterconfig):
        """Prepare the master aspect of job submission.
           Returns: jobmasterconfig understood by Panda backend."""

        job = app._getParent()
        logger.debug('ProdTransPandaRTHandler master_prepare() for %s',
                    job.getFQID('.'))

        return None

    def prepare(self, app, appsubconfig, appmasterconfig, jobmasterconfig):
        """Prepare the specific aspec of each subjob.
           Returns: subjobconfig list of objects understood by backends."""

        from pandatools import Client
        from pandatools import AthenaUtils
        from taskbuffer.JobSpec import JobSpec
        from taskbuffer.FileSpec import FileSpec
        from GangaAtlas.Lib.ATLASDataset.DQ2Dataset import dq2_set_dataset_lifetime
        from GangaPanda.Lib.Panda.Panda import refreshPandaSpecs
        
        # make sure we have the correct siteType
        refreshPandaSpecs()

        job = app._getParent()
        masterjob = job._getRoot()

        logger.debug('ProdTransPandaRTHandler prepare called for %s',
                     job.getFQID('.'))

        job.backend.actualCE = job.backend.site
        job.backend.requirements.cloud = Client.PandaSites[job.backend.site]['cloud']

        # check that the site is in a submit-able status
        if not job.splitter or job.splitter._name != 'DQ2JobSplitter':
            allowed_sites = job.backend.list_ddm_sites()

        try:
            outDsLocation = Client.PandaSites[job.backend.site]['ddm']
            tmpDsExist = False
            if (configPanda['processingType'].startswith('gangarobot') or configPanda['processingType'].startswith('hammercloud')):
                #if Client.getDatasets(job.outputdata.datasetname):
                if getDatasets(job.outputdata.datasetname):
                    tmpDsExist = True
                    logger.info('Re-using output dataset %s'%job.outputdata.datasetname)
            if not configPanda['specialHandling']=='ddm:rucio' and not  configPanda['processingType'].startswith('gangarobot') and not configPanda['processingType'].startswith('hammercloud') and not configPanda['processingType'].startswith('rucio_test'):
                Client.addDataset(job.outputdata.datasetname,False,location=outDsLocation,allowProdDisk=True,dsExist=tmpDsExist)
            logger.info('Output dataset %s registered at %s'%(job.outputdata.datasetname,outDsLocation))
            dq2_set_dataset_lifetime(job.outputdata.datasetname, outDsLocation)
        except exceptions.SystemExit:
            raise BackendError('Panda','Exception in adding dataset %s: %s %s'%(job.outputdata.datasetname,sys.exc_info()[0],sys.exc_info()[1]))
        
        # JobSpec.
        jspec = JobSpec()
        jspec.currentPriority = app.priority
        jspec.jobDefinitionID = masterjob.id
        jspec.jobName = commands.getoutput('uuidgen 2> /dev/null')
        jspec.coreCount = app.core_count
        jspec.AtlasRelease = 'Atlas-%s' % app.atlas_release
        jspec.homepackage = app.home_package
        jspec.transformation = app.transformation
        jspec.destinationDBlock = job.outputdata.datasetname
        if job.outputdata.location:
            jspec.destinationSE = job.outputdata.location
        else:
            jspec.destinationSE = job.backend.site
        if job.inputdata:
            jspec.prodDBlock = job.inputdata.dataset[0]
        else:
            jspec.prodDBlock = 'NULL'
        if app.prod_source_label:
            jspec.prodSourceLabel = app.prod_source_label
        else:
            jspec.prodSourceLabel = configPanda['prodSourceLabelRun']
        jspec.processingType = configPanda['processingType']
        jspec.specialHandling = configPanda['specialHandling']
        jspec.computingSite = job.backend.site
        jspec.cloud = job.backend.requirements.cloud
        jspec.cmtConfig = app.atlas_cmtconfig
        if app.dbrelease == 'LATEST':
            try:
                latest_dbrelease = getLatestDBReleaseCaching()
            except:
                from pandatools import Client
                latest_dbrelease = Client.getLatestDBRelease()
            m = re.search('(.*):DBRelease-(.*)\.tar\.gz', latest_dbrelease)
            if m:
                self.dbrelease_dataset = m.group(1)
                self.dbrelease = m.group(2)
            else:
                raise ApplicationConfigurationError(None, "Error retrieving LATEST DBRelease. Try setting application.dbrelease manually.")
        else:
            self.dbrelease_dataset = app.dbrelease_dataset
            self.dbrelease = app.dbrelease
        jspec.jobParameters = app.job_parameters

        if self.dbrelease:
            if self.dbrelease == 'current':
                jspec.jobParameters += ' --DBRelease=current' 
            else:
                if jspec.transformation.endswith("_tf.py") or jspec.transformation.endswith("_tf"):
                    jspec.jobParameters += ' --DBRelease=DBRelease-%s.tar.gz' % (self.dbrelease,)
                else:
                    jspec.jobParameters += ' DBRelease=DBRelease-%s.tar.gz' % (self.dbrelease,)
                dbspec = FileSpec()
                dbspec.lfn = 'DBRelease-%s.tar.gz' % self.dbrelease
                dbspec.dataset = self.dbrelease_dataset
                dbspec.prodDBlock = jspec.prodDBlock
                dbspec.type = 'input'
                jspec.addFile(dbspec)

        if job.inputdata:
            m = re.search('(.*)\.(.*)\.(.*)\.(.*)\.(.*)\.(.*)',
                          job.inputdata.dataset[0])
            if not m:
                logger.error("Error retrieving run number from dataset name")
                #raise ApplicationConfigurationError(None, "Error retrieving run number from dataset name")
                runnumber = 105200
            else:
                runnumber = int(m.group(2))
            if jspec.transformation.endswith("_tf.py") or jspec.transformation.endswith("_tf"):
                jspec.jobParameters += ' --runNumber %d' % runnumber
            else:
                jspec.jobParameters += ' RunNumber=%d' % runnumber
        
        # Output files.
        randomized_lfns = []
        ilfn = 0
        for lfn, lfntype in zip(app.output_files,app.output_type):
            ofspec = FileSpec()
            if app.randomize_lfns:
                randomized_lfn = lfn + ('.%s.%d.%s' % (job.backend.site, int(time.time()), commands.getoutput('uuidgen 2> /dev/null')[:4] ) )
            else:
                randomized_lfn = lfn
            ofspec.lfn = randomized_lfn
            randomized_lfns.append(randomized_lfn)
            ofspec.destinationDBlock = jspec.destinationDBlock
            ofspec.destinationSE = jspec.destinationSE
            ofspec.dataset = jspec.destinationDBlock
            ofspec.type = 'output'
            jspec.addFile(ofspec)
            if jspec.transformation.endswith("_tf.py") or jspec.transformation.endswith("_tf"):
                jspec.jobParameters += ' --output%sFile %s' % (lfntype, randomized_lfns[ilfn])
            else:
                jspec.jobParameters += ' output%sFile=%s' % (lfntype, randomized_lfns[ilfn])
            ilfn=ilfn+1

        # Input files.
        if job.inputdata:
            for guid, lfn, size, checksum, scope in zip(job.inputdata.guids, job.inputdata.names, job.inputdata.sizes, job.inputdata.checksums, job.inputdata.scopes):
                ifspec = FileSpec()
                ifspec.lfn = lfn
                ifspec.GUID = guid
                ifspec.fsize = size
                ifspec.md5sum = checksum
                ifspec.scope = scope
                ifspec.dataset = jspec.prodDBlock
                ifspec.prodDBlock = jspec.prodDBlock
                ifspec.type = 'input'
                jspec.addFile(ifspec)
            if app.input_type:
                itype = app.input_type
            else:
                itype = m.group(5)
            if jspec.transformation.endswith("_tf.py") or jspec.transformation.endswith("_tf"):
                jspec.jobParameters += ' --input%sFile %s' % (itype, ','.join(job.inputdata.names))
            else:
                jspec.jobParameters += ' input%sFile=%s' % (itype, ','.join(job.inputdata.names))

        # Log files.
        lfspec = FileSpec()
        lfspec.lfn = '%s.job.log.tgz' % jspec.jobName
        lfspec.destinationDBlock = jspec.destinationDBlock
        lfspec.destinationSE  = jspec.destinationSE
        lfspec.dataset = jspec.destinationDBlock
        lfspec.type = 'log'
        jspec.addFile(lfspec)
        
        return jspec

from Ganga.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers
allHandlers.add('ProdTrans', 'Panda', ProdTransPandaRTHandler)



##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: IApplication.py,v 1.1 2008-07-17 16:40:52 moscicki Exp $
##########################################################################

from Ganga.GPIDev.Adapters.IApplication import IApplication
from Ganga.Core.GangaRepository import getRegistry
from Ganga.GPIDev.Base.Proxy import GPIProxyObjectFactory, isType
from Ganga.GPIDev.Lib.GangaList.GangaList import GangaList
from Ganga.GPIDev.Lib.File import File
import os
import shutil
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem

import Ganga.Utility.logging
logger = Ganga.Utility.logging.getLogger()


from Ganga.Utility.Config import getConfig
from Ganga.Utility.files import expandfilename

config = getConfig("Preparable")

class IPrepareApp(IApplication):

    """
    Base class for all applications which can be placed into a prepared\
    state. 
    """
    _schema = Schema(Version(0, 0), {'hash': SimpleItem(
        defvalue=None, typelist=['type(None)', 'str'], hidden=1)})
    _category = 'applications'
    _name = 'PrepareApp'
    _hidden = 1

    def __init__(self):
        super(IPrepareApp, self).__init__()

    def _auto__init__(self, unprepare=None):
        if unprepare is True:
            logger.debug("Calling unprepare() from IPrepareApp's _auto__init__()")
            self.unprepare()

    def prepare(self, force=False):
        """
        Base class for all applications which can be placed into a prepared\
        state. 

        """
        pass

    def post_prepare(self):
        """
        Put any methods that should always be run at the end of the preparation process here.
        """
        self.calc_hash()

    def unprepare(self, force=False):
        """
        Revert an application back to the exact state it was in prior to being\
        prepared.
        """
        logger.debug("Running unprepare() from IPrepareApp")
        if self.is_prepared is True:
            self.is_prepared = None
        elif self.is_prepared is not None:
            self.is_prepared = None
        self.hash = None

    def copyPreparables(self):
        """
        This method iterates over all attributes in an application and decides\
        whether they should be persisted (i.e. copied) in the Shared Directory\
        when the application is prepared.
        If an IOError is raised when attempting the copy, this method returns 0\
        otherwise it returns 1.
        """
        send_to_sharedir = []
        for name, item in self._schema.allItems():
            if item['preparable']:
                logger.debug('Found preparable %s' % (name))
                logger.debug('adding to sharedir %s' % (self.__getattribute__(name)))
                send_to_sharedir.append(self.__getattribute__(name))
        shared_path = os.path.join(expandfilename(getConfig('Configuration')['gangadir']), 'shared', getConfig('Configuration')['user'])
        for prepitem in send_to_sharedir:
            logger.debug('working on %s' % (prepitem))
            # we may have a list of files/strings
            if isType(prepitem, (list, GangaList)):
                logger.debug('found a list')
                for subitem in prepitem:
                    if isType(subitem, str):
                        # we have a file. if it's an absolute path, copy it to
                        # the shared dir
                        if os.path.abspath(subitem) == subitem:
                            logger.debug('Sending file %s to shared directory.' % (subitem))
                            try:
                                shr_dir = os.path.join(shared_path, self.is_prepared.name)
                                if not os.path.isidr(shr_dir):
                                    os.makedirs(shr_dir)
                                shutil.copy2(subitem, shr_dir)
                                logger.debug("Copying into: %s" % str(shr_dir))
                            except IOError as e:
                                logger.error(e)
                                return 0
                    elif isType(subitem, File) and subitem.name is not '':
                        logger.debug('Sending file object %s to shared directory' % subitem.name)
                        try:
                            shr_dir = os.path.join(shared_path, self.is_prepared.name)
                            if not os.path.isdir(shr_dir):
                                os.makedirs(shr_dir)
                                logger.debug("Copying into: %s" % str(shr_dir))
                            shutil.copy2(subitem.name, shr_dir)
                        except IOError as e:
                            logger.error(e)
                            return 0
            elif isinstance(prepitem, str):
                logger.debug('found a string')
                # we have a file. if it's an absolute path, copy it to the
                # shared dir
                if os.path.abspath(prepitem) == prepitem:
                    logger.debug('Sending string file %s to shared directory.' % (prepitem))
                    try:
                        shr_dir = os.path.join(shared_path, self.is_prepared.name)
                        if not os.path.isdir(shr_dir):
                            os.makedirs(shr_dir)
                        logger.debug("Copying into: %s" % str(shr_dir))
                        shutil.copy2(prepitem, shr_dir)
                    except IOError as e:
                        logger.error(e)
                        return 0
            elif isType(prepitem, File) and prepitem.name is not '':
                logger.debug('found a file')
                logger.debug('Sending "File" object %s to shared directory' % prepitem.name)
                try:
                    shr_dir = os.path.join(shared_path, self.is_prepared.name)
                    if not os.path.isdir(shr_dir):
                        os.makedirs(shr_dir)
                    logger.debug("Copying into: %s" % str(shr_dir))
                    shutil.copy2(prepitem.name, shr_dir)
                except IOError as e:
                    logger.error(e)
                    return 0
            else:
                logger.debug('Nothing worth copying found in %s' % (prepitem))
        return 1

    def calc_hash(self, verify=False):
        """Calculate the MD5 digest of the application's preparable attribute(s), and store
        that value in the application schema. The value is recalculated (and compared against
        the initial value) every time the application is written to the Ganga repository. This
        allows warnings to be generated should an application's locked attributes be changed 
        post-preparation.
        """
        from Ganga.GPIDev.Base.Proxy import runProxyMethod
        import cStringIO
        try:
            import hashlib
            digest = hashlib.new('md5')
        except Exception as err:
            logger.debug("Err: %s" % str(err))
            import md5
            digest = md5.new()

        sio = cStringIO.StringIO()
        runProxyMethod(self, 'printPrepTree', sio)
        digest.update(str(sio.getvalue()))
        tmp = sio.getvalue()
        if verify == False:
            self.hash = digest.hexdigest()
        else:
            # we return true if this is called with verify=True and the current hash is the same as that stored in the schema.
            # this is checked immediately prior to (re)writing the object to
            # the repository
            return digest.hexdigest() == self.hash

    def incrementShareCounter(self, shared_directory_name):
        logger.debug('Incrementing shared directory reference counter')
        shareref = GPIProxyObjectFactory(getRegistry("prep").getShareRef())
        logger.debug('within incrementShareCounter, calling increase')
        shareref.increase(shared_directory_name)

    def decrementShareCounter(self, shared_directory_name, remove=0):
        remove = remove
        logger.debug('Decrementing shared directory reference counter')
        shareref = GPIProxyObjectFactory(getRegistry("prep").getShareRef())
        shareref.decrease(shared_directory_name, remove)

    def listShareDirs(self):
        shareref = GPIProxyObjectFactory(getRegistry("prep").getShareRef())
        shareref

    def listShareDirContents(self, shared_directory_name):
        shareref = GPIProxyObjectFactory(getRegistry("prep").getShareRef())
        shareref.ls(shared_directory_name)

    def checkPreparedHasParent(self, prepared_object):
        if prepared_object._getRegistry() is None:
            self.incrementShareCounter(prepared_object.is_prepared.name)
            self.decrementShareCounter(prepared_object.is_prepared.name)
            logger.info('Application is not currently associated with a persisted Ganga object')
            logger.info('(e.g. box, job, task). Both the prepared application and the contents of')
            logger.info('its shared directory will be lost when Ganga exits.')
            logger.info('Shared directory location: %s' % (self.is_prepared.name))
            # logger.error(self.listShareDirContents(prepared_object.is_prepared.name))
        else:
            self.incrementShareCounter(prepared_object.is_prepared.name)



from Ganga.Utility.Config import getConfig
from Ganga.Utility.logging import getLogger

config = getConfig('Tasks')

logger = getLogger()

#  Helper function for singular/plural


def say(number, unit):
    if number == 1:
        return "one %s" % (unit)
    else:
        return "%s %ss" % (number, unit)
